///////////////////////////////////////////////////
// Ramsey On-line game
// Author: Przemek Gordinowicz 
///////////////////////////////////////////////////
// Program creates script R_*.sh that runs the main program "Ramsey"
///////////////////////////////////////////////////
// Parameters: (option) prefix symmetric 
//            without option - merge and run
//            -m             - only merge
//            -e             - only shorten - removing existing configurations 
//            -d             - only description
//            -de            - only description with using short files 
//            -r nr          - only run (with partitioning into nr of pieces)
//            -rd nr         - run on files with descriptions (with partitioning into nr of pieces)
//            -a             - "amnesty" 
//            -u p_new       - "go up" - existing files will be names as run ones 
//                        with the number of allowed edges increased by 1 (it allows finishing 
//                        the algorithm with a smaller bound (after removing some files) 
//                        the name will contain a new prefix
///////////////////////////////////////////////////
// Browse the file prefiks.txt for info about files with a particular prefix (encoding number of edges)
// written by Ramsey. Comparing number of coloured edges creates batch file with list of calls  
// for the next iteration of Ramsey program according to:
//    r_1*b_2 > 0 && max(r_1-1, max_r_2) + max(b_1, b_2-1) < edge_no
// or
//    b_1*r_2 > 0 && max(r_1, max_r_2-1) + max(b_1-1, b_2) < edge_no
// and/or analogous when the symmetric Ramsey number is computed 
//////////////////////////////////////////////////
// Note: number of edges lower than 255!
/////////////////////////////////////////////////
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <math.h>
#include "ramsey.h"

void quicksort(long *buffer, long d, long g) {
  // sorting buffer[d..g-1]
  if (g - d > 1) {
    long i, s;
    long t, t2;
    
    // change the first and the middle (not random!)
    i = (d+g)/2;
    t = buffer[i];
    buffer[i] = buffer[d];
    buffer[d] = t;

    s = d+1;
    while(s < g && buffer[s] < t) s++;
    for (i = s--; i < g; i++)
      if (buffer[i] < t) {
        t2 = buffer[++s];
        buffer[s] = buffer[i];
        buffer[i] = t2;
      }
    buffer[d] = buffer[s];
    buffer[s] = t;
    // now buffer[d..s-1] < buffer[s] < buffer[s+1..g-1]
      
    quicksort(buffer, d, s); //sort [d..s-1]
    quicksort(buffer, s+1, g); //sort [s+1..g-1]
  }
}

int max(int a, int b) {
  return (a > b ? a : b);
}

int main(int argc, char * argv[]) {
  if (argc >= 2) {
      FILE *report;
      long liczby[20000];
      double fsize[20000];
      int max_kraw;
      int ile, i, j, n, m, w, x, y, z;
      double total_size, size, c_size, t;
      struct stat descript;
      int k, r, b;
      int _r, _b;
      int merge, run, split, go_up, descr, shorten, amnesty;
      int good, symmetric;
      char *prefix, *p_new;
      char buf[200];
      char fname[50];
      char prefix2[20];

      // parse args 
      if (*argv[1] == '-') {
        merge = argv[1][1] == 'm';
        run = argv[1][1] == 'r';
        go_up = argv[1][1] == 'u';
        shorten = argv[1][1] == 'e';
        descr = argv[1][1] == 'd';
        amnesty = argv[1][1] == 'a';
        i = 2;
        if (go_up) p_new = argv[i++];
        if (descr) {
          shorten = argv[1][2] == 'e';
        }
        if (run) {
          descr = argv[1][2] == 'd';
          shorten = argv[1][2] == 'e';
          split = atol(argv[i++]);
        }
      }
      else {
        i = 1;
        merge = run = split = 1;
        go_up = descr = shorten = 0;
      }
      prefix = argv[i++];
      symmetric = argc > i && atol(argv[i++]) == 1;

      if (go_up) {
        // change report file and filenames allowing one edge more (for tests) 
        FILE *report_new;         
        sprintf(buf, "%s.txt", prefix);
        report = fopen(buf, "rt");
        sprintf(buf, "%s.txt", p_new);
        report_new = fopen(buf, "wt");
        while(!feof(report)) {
          fscanf(report, "%d_%d_%d\n", &k, &r, &b);
          fprintf(report_new, "%d_%d_%d\n", k+1, r, b);
          sprintf(buf, "%s %s_%d_%d_%d %s_%d_%d_%d", mv_command, prefix, k, r, b, p_new, k+1, r, b);
          system(buf);
        }
        fclose(report_new);
        fclose(report);
      }
      else {
        if (merge) {
          // at first sorting to remove repetitions 
          sprintf(buf, "%s %s.txt %s.tx", mv_command, prefix, prefix);
          system(buf);
          sprintf(buf, "%s -u %s.tx >%s.txt", sort_command, prefix, prefix);
          system(buf);
          sprintf(buf, "%s %s.tx", del_command, prefix);
          system(buf);
        }

        sprintf(buf, "%s.txt", prefix);
        report = fopen(buf, "rt");
        for(ile = 0; !feof(report); ile++) {
          fscanf(report, "%d_%d_%d\n", &k, &r, &b);
          liczby[ile] = (k << 16) + (r << 8) + b;
        }
        fclose(report);
        quicksort(liczby, 0, ile);

        max_kraw = liczby[0] >> 16;

        if (merge) {
          sprintf(buf, "%s_%dm.%s", prefix, max_kraw, batch_extension);
          FILE *batch= fopen(buf, "wt");      
          // merge
          for (i = 0; (i < ile) && (liczby[i] >> 16 == max_kraw); i++) {
             _b = liczby[i] & 255;
             _r = (liczby[i] >> 8) & 255;

             fprintf(batch, "%s %s_%d_%d_%d\n", merge_command, prefix, max_kraw, _r, _b);
          }
          fclose(batch);
        }
        if (amnesty) {
          sprintf(buf, "%s_%da.%s", prefix, max_kraw, batch_extension);
          FILE *batch= fopen(buf, "wt");      
          // descript
          for (i = 0; i < ile; i++) {
             _b = liczby[i] & 255;
             _r = (liczby[i] >> 8) & 255;
             k = liczby[i] >> 16;

             fprintf(batch, "%s -a %s_%d_%d_%d %d\n", ramsey_command, prefix, k, _r, _b, symmetric);
          }
          fclose(batch);
        }
        if (shorten && !descr && !run) {
          sprintf(buf, "%s_%de.%s", prefix, max_kraw, batch_extension);
          FILE *batch= fopen(buf, "wt");      
  
          for (i = 0; (i < ile) && (liczby[i] >> 16 == max_kraw); i++) {
            _b = liczby[i] & 255;
            _r = (liczby[i] >> 8) & 255;

            fprintf(batch, "%s -e %s %s_%d_%d_%d %d -n %s_%d_%d_%d", ramsey_command, prefix, prefix, max_kraw, _r, _b, symmetric, prefix, max_kraw, _r, _b);

            for (j = i+1; j < ile; j++) {
              b = liczby[j] & 255;
              r = (liczby[j] >> 8) & 255;
              k = liczby[j] >> 16;
              if (b == _b && r == _r) fprintf(batch, " %s_%d_%d_%d", prefix, k, r, b);
            }
            fprintf(batch, " -r");
            for (j = i-1; j >= 0; j--) {
              b = liczby[j] & 255;
              r = (liczby[j] >> 8) & 255;
              if (b == _b && r == _r-1) { 
                fprintf(batch, " %s_%d_%d_%d", prefix, max_kraw, r, b);
                break;
              } else if (r < _r-1 || (r == _r-1 && b < _b)) break;
            }
            for (j = i+1; j < ile; j++) {
              b = liczby[j] & 255;
              r = (liczby[j] >> 8) & 255;
              k = liczby[j] >> 16;
              if (b == _b && r == _r-1) fprintf(batch, " %s_%d_%d_%d", prefix, k, r, b);
            }
            fprintf(batch, " -b");
            if (i > 0 && liczby[i-1] == liczby[i]-1) fprintf(batch, " %s_%d_%d_%d", prefix, max_kraw, _r, _b-1);
            for (j = i+1; j < ile; j++) {
              b = liczby[j] & 255;
              r = (liczby[j] >> 8) & 255;
              k = liczby[j] >> 16;
              if (b == _b-1 && r == _r) fprintf(batch, " %s_%d_%d_%d", prefix, k, r, b);
            }
            fprintf(batch, "\n");
          }
          fclose(batch);
        }

        if (descr && !run) {
          if (shorten)
            sprintf(prefix2, "%s%s", prefix, short_extension);
          else strcpy(prefix2, prefix);
          sprintf(buf, "%s_%dd.%s", prefix, max_kraw, batch_extension);
          FILE *batch= fopen(buf, "wt");      
          // descript
          for (i = 0; (i < ile) && (liczby[i] >> 16 == max_kraw); i++) {
             _b = liczby[i] & 255;
             _r = (liczby[i] >> 8) & 255;

             fprintf(batch, "%s %s %s_%d_%d_%d\n", ramsey_command, prefix, prefix2, max_kraw, _r, _b);
          }
          fclose(batch);
        }

        if (run) {   
          if (descr)
            sprintf(prefix2, "%s%s", prefix, descr_extension);
          else if (shorten)
            sprintf(prefix2, "%s%s", prefix, short_extension);
          else strcpy(prefix2, prefix);
          //rozmiary plików
          for (i =0; i < ile; i++) {
            b = liczby[i] & 255;
            r = (liczby[i] >> 8) & 255;
            k = liczby[i] >> 16;
            sprintf(buf, "%s_%d_%d_%d", prefix2, k, r, b);
            if (stat(buf, &descript) == 0)
              fsize[i] = ((double)descript.st_size)/1024; //file size in KB
          }

          sprintf(buf, "%s_%d.%s", prefix, max_kraw, batch_extension);
          FILE *batch= fopen(buf, "wt");
          m = 0;
          total_size = 0;
          for (i = 0; (i < ile) && (liczby[i] >> 16 == max_kraw); i++) {
             _b = liczby[i] & 255;
             _r = (liczby[i] >> 8) & 255;

             for (j = i; j < ile; j++) {
               b = liczby[j] & 255;
               r = (liczby[j] >> 8) & 255;
               k = liczby[j] >> 16;

               good = (_r*b > 0 && max(r, _r-1) + max(b-1, _b) < max_kraw) ||
                      (r*_b > 0 && max(r-1, _r) + max(b, _b-1) < max_kraw);

               if (symmetric)
                 good = good || (_r*r > 0 && max(b, _r-1) + max(r-1, _b) < max_kraw) ||
                                (_b*b > 0 && max(b-1, _r) + max(r, _b-1) < max_kraw);
               if (good) {
                  m++;
                  total_size += fsize[i] * fsize[j];
                  fprintf(batch, "%s %d %s %d %s_%d_%d_%d %s_%d_%d_%d %d\n", ramsey_command, m, prefix, max_kraw-1, prefix2, max_kraw, _r, _b, prefix2, k, r, b, symmetric);
               }
             }
          }
          fclose(batch);
          printf("Ramrun %d. Tasks: %d. Total size: %.1f\n", max_kraw, m, total_size);
          if (split > 1) { // divide the work into "split" pieces (for paralell computing)  
            m = 0; i = 0; j = 0;
            n = 0; 
            c_size = 0;
            size = total_size/(split-n); //size +/- of task in n-th script
            sprintf(buf, "%s_%d_%d.%s", prefix, max_kraw, n, batch_extension);
            FILE *batch= fopen(buf, "wt");
            
            for (i = 0; (i < ile) && (liczby[i] >> 16 == max_kraw); i++) {
               _b = liczby[i] & 255;
               _r = (liczby[i] >> 8) & 255;

               for (j = i; j < ile; j++) {
                 b = liczby[j] & 255;
                 r = (liczby[j] >> 8) & 255;
                 k = liczby[j] >> 16;

                 good = (_r*b > 0 && max(r, _r-1) + max(b-1, _b) < max_kraw) ||
                        (r*_b > 0 && max(r-1, _r) + max(b, _b-1) < max_kraw);

                 if (symmetric)
                   good = good || (_r*r > 0 && max(b, _r-1) + max(r-1, _b) < max_kraw) ||
                                  (_b*b > 0 && max(b-1, _r) + max(r, _b-1) < max_kraw);
                 if (good) {
                    if (size < 3*fsize[i] * fsize[j]) {  // each file has at least 3 tasks    
                      if (2*fsize[i] > sqrt(size)) {  // divide i-th file (when to big)
                        // split file[i], into x +/- equal pieces 
                        x = (int)ceil(3*fsize[i]/sqrt(size)); // when file need to be divided it is split into more pieces than necesary
                        sprintf(fname, "%s_%d_%d_%d", prefix2, max_kraw, _r, _b);
                        sprintf(buf, "%s_0", fname);
                        if (stat(buf, &descript)) { // when file wasn't split, split it now 
                          sprintf(buf, "%s -d -C %d %s %s_", split_command, (int)ceil(1024*fsize[i]/x+120), fname, fname);
                          system(buf);

                          z = x > 10 ? 9 : x - 1;  // onedigit names
                          for ( ; z >= 0; z--) {
                            sprintf(buf, "%s %s_0%d %s_%d", mv_command, fname, z, fname, z);
                            system(buf);
                          }
                          printf("%d %s split into %d pieces.\n", m+1, fname, x);
                        }
                        else x = (int)ceil(fsize[i]*1024/descript.st_size);
                      }
                      else x = 1;
                      if (i != j && 2*fsize[j] > sqrt(size)) { //split j-th file 
                        // split file[i], into x +/- equal pieces
                        y = (int)ceil(3*fsize[j]/sqrt(size)); // when file need to be divided it is split into more pieces
                        sprintf(fname, "%s_%d_%d_%d", prefix2, k, r, b);
                        sprintf(buf, "%s_0", fname);
                        if (stat(buf, &descript)) { // when file wasn't split, split it now
                          sprintf(buf, "%s -d -C %d %s %s_", split_command, (int)ceil(1024*fsize[j]/y+120), fname, fname);
                          system(buf);

                          z = y > 10 ? 9 : y - 1;  // onedigit names
                          for ( ; z >= 0; z--) {
                            sprintf(buf, "%s %s_0%d %s_%d", mv_command, fname, z, fname, z);
                            system(buf);
                          }
                          printf("%d %s split into %d pieces.\n", m+1, fname, y);
                        }
                        else y = (int)ceil(fsize[j]*1024/descript.st_size);
                      }
                      else y = 1;
                      if (i == j) { // the same pair - x files
                        t = fsize[i]*fsize[i]/x/x; // roughly size of one piece 
                        for (y = 0; y < x; y++) {
                          total_size -= y * t;  
                          for (z = y; z < x; z++) {
                            m++;
                            if (fabs(size-c_size) < fabs(size - c_size - t)) { // new fil
                              fclose(batch);
                              total_size -= c_size;        // size of the rest 
                              printf("File no %d, up to: %d, work-size: %.1f\n", n, m-1, c_size);
                              n++; c_size = 0;
                              size = total_size/(split-n); //size +/- of a task in n-th script 
                              sprintf(buf, "%s_%d_%d.%s", prefix, max_kraw, n, batch_extension);
                              FILE *batch= fopen(buf, "wt");
                            }
                            c_size += t;
                            fprintf(batch, "%s %d %s %d %s_%d %s_%d %d\n", ramsey_command, m, prefix, max_kraw-1, fname, y, fname, z, symmetric);
                          }
                        }
                      }
                      else { // x -of i-th, y -of j-th files
                        t = fsize[i]/x * fsize[j]/y;
                        for (w = 0; w < x; w++)
                          for (z = 0; z < y; z++) {
                            m++;
                            if (fabs(size-c_size) < fabs(size - c_size - t)) { // new file
                              fclose(batch);
                              total_size -= c_size;        // size of the rest
                              printf("File no %d, up to: %d, work-size: %.1f\n", n, m-1, c_size);
                              n++; c_size = 0;
                              size = total_size/(split-n); //size +/- of a task in n-th script
                              sprintf(buf, "%s_%d_%d.%s", prefix, max_kraw, n, batch_extension);
                              FILE *batch= fopen(buf, "wt");
                            }
                            c_size += t;
                            if (x == 1) fprintf(batch, "%s %d %s %d %s_%d_%d_%d %s_%d %d\n", ramsey_command, m, prefix, max_kraw-1, prefix2, max_kraw, _r, _b, fname, z, symmetric);
                            else if (y == 1) fprintf(batch, "%s %d %s %d %s_%d %s_%d_%d_%d %d\n", ramsey_command, m, prefix, max_kraw-1, fname, w, prefix2, k, r, b, symmetric);
                            else fprintf(batch, "%s %d %s %d %s_%d_%d_%d_%d %s_%d %d\n", ramsey_command, m, prefix, max_kraw-1, prefix2, max_kraw, _r, _b, w, fname, z, symmetric);
                          }
                      }
                    }
                    else {
                      m++;
                      if (fabs(size-c_size) < fabs(size - c_size - fsize[i]*fsize[j])) { // nowy plik
                        fclose(batch);
                        total_size -= c_size;       // size of the rest
                        printf("File no %d, up to: %d, work-size: %.1f\n", n, m-1, c_size);
                        n++; c_size = 0;
                        size = total_size/(split-n); //size +/- of a task in n-th script
                        sprintf(buf, "%s_%d_%d.%s", prefix, max_kraw, n, batch_extension);
                        FILE *batch= fopen(buf, "wt");
                      }
                      c_size += fsize[i] * fsize[j];
                      fprintf(batch, "%s %d %s %d %s_%d_%d_%d %s_%d_%d_%d %d\n", ramsey_command, m, prefix, max_kraw-1, prefix2, max_kraw, _r, _b, prefix2, k, r, b, symmetric);
                    }
                 }
               }
            }
            fclose(batch);
            printf("File no %d, up to: %d, work-size: %.1f\n", n, m, c_size);
          }
        }
      }
  }
  return 0;
}
