///////////////////////////////////////////////////
// Ramsey On-line game
// Author: Przemek Gordinowicz 
//         Pawel Pralat 
///////////////////////////////////////////////////
//    Main tool - combines and modifies files with configurations
//    The main algorithm implemented in 2 functions called combine() with different parameters
//    (file ramsey_alg.cpp)
//
//    Filenames:
//    prefix_maxedges_redcount_bluecount (eg. RK3K4_14_4_3)
//
//    Parameters (3 base variants): 
//    (1)  file_name - prints colured graphs from a file 
//
//    (4)  prefix maxedges redgraph bluegraph
//    creates the last layer writing to files:
//                        prefix_maxedges_redcount_0
//                        prefix_maxedges_0_bluecount
//    corresponding monochromatic graphs (redgraph and bluegraph are filenames with graphs)
//
//    (5 or 6) processno prefix maxedges fname1 fname2 [sym]
//    one iteration of the algorithm
//    read from files: prefix_maxedges1_redcount1_bluecount1
//                     prefix_maxedges2_redcount2_bluecount2
//    write to files:  prefix_maxedges_redcount_bluecount_processno
//    using temporary files: prefix_maxedges_recount_bluecount_processno_nextno
//      
//    if sym is set to 1 we have symmetric case (the same red and blue base graph) 
//      - we choose resulting graph (exchanging colours) such that:
//         - red edges is not less than blue ones
//         - if numbers are the same we choose smaller (lexically) canonical form   
//
//    Additional variants:
//    -e prefix filename1 [symmetric] -n {filenames ...} -r {filenames ...} -b {filenames ...} 
//    remove superconfigurations of already considered (take configurations from filename1) and check
//    whether one of list of files contains its subconfiguration result is written to a file with name
//    similar to filename with "s" add (after prefix)
//    (See the second last parapgraph of Section 2)
//
//    -a filename [symmetric]
//     "amnesty"/release of edges of configuration marked as NON_EDGE to status NON_DETERMINED
//     it works with exchanging program file ramsey into the second version (see ramsey/readme.txt 
//     for script that use it) (See also the last paragraph of Section 2) 
//
//    (2) prefix filename
//    creates "description" - list of coloured edges that can be distinguish without creating isomorphic copy 
//    result written to a file which name has "d" add (after prefix)
//    (See the third last paragraph of Section 2)
//
#include "gtools.h"
#include "ramsey.h"
#include <stdio.h>
#include <stdlib.h>

// Remark: definition moved to makefile
// If written 0 instead of 1 then non_edge and not_determined has the same meaning
// #define NON_EDGE 1

const int nr_colours = 2;
// number of layers = ceil(log_2(nr_colors+2))
const int levels = 2; 
// size of a node in dictionary tree 
const int node_size = 8 * sizeof(void *);

int max_edges, symmetric;
//diagnostics - couters
long est1_count, est2_count;  // global variables (0's at start)
long est1_stop, est2_stop;
long graph_create, graph_stop;
long graph_create2, graph_stop2;

char *fname_prefix, *fname_suffix;
char *description, *pg_readg_line;

int max_memory = 1024*1024*768;
int mem_avail;

void *graph_dict;
int *files_count;

// global, reused to not waste time for allocation of memory 
// assumed that numbers of vertices and edges are lower than 100
int used[100];
int assign[100];

#include "pg_nauty.c"

int max(int a, int b) {
  return (a > b ? a : b);
}

int min(int a, int b) {
  return (a > b ? b : a);
}

#include "ramsey_graph.cpp"
#include "ramsey_dict.cpp"
#include "ramsey_aux.cpp"
#include "ramsey_alg.cpp"

///////////////////////////////////////////////////
// Main program
///////////////////////////////////////////////////
int main(int argc, char * argv[])
{
  for (int i = 1; i < argc; i++)
    fprintf(stderr, "%s ", argv[i]);

  mem_avail = max_memory;
  if (argv[1][0] == '-' && argv[1][1] == 'e') { // remove superconfigurations of already considered 
    FILE *in, *in2, *out;                              
    graph *g;
    char fname[50];
    char command[100];
    int i, j, n, m, kr, kb, nr = 0;
    col_graph *cg;

    graph_dict = malloc((nr_colours+2)*sizeof(void*));
    memset(graph_dict, 0, (nr_colours+2)*sizeof(void*));
    
    i = 4;
    if (*argv[i] != '-') symmetric = (atol(argv[i++]) == 1);
    i++;
    kr = 0;  
    kb = 0; 
    for (j = 1; i < argc; i++)
      if (*argv[i] == '-') j++;
      else {
        if (!strcmp(argv[i], argv[3])) j--; 
        in = fopen(argv[i], "rt");
        while (pg_readg_line = pg_getline(in)) {
          inserttext(((void**)graph_dict)+j, pg_readg_line, -1);
          if (mem_avail < 0) { 
            if (nr) sprintf(fname, "%s%s%s_%d", argv[2], short_extension, strchr(argv[3], '_'), nr-1); 
            else strcpy(fname, argv[3]);
            in2 = fopen(fname, "rt");
            sprintf(fname, "%s%s%s_%d", argv[2], short_extension, strchr(argv[3], '_'), nr); 
            out = fopen(fname, "wt");
            while (TRUE) {
              if ((g = pg_readg(in2,NULL,0,&m,&n)) == NULL) break;
              cg = new col_graph(g, n);
              graph_create++;
              if (check_excess(cg, kr, kb))
                fprintf(out, "%s\n", pg_readg_line);
              else graph_stop++;
              delete cg;
              FREES(g);
            }
            fclose(out);
            fclose(in2);
            for (n = 0; n < nr_colours+2; n++)
              if (((void**)graph_dict)[n]) free_dict(((void**)graph_dict)[n]);            
            memset(graph_dict, 0, (nr_colours+2)*sizeof(void*));
            nr++;
            kr = kb = 0; //not need anymore
          }
        }
        fclose(in);
        if (!strcmp(argv[i], argv[3])) j++; 
        if (mem_avail < 0) break; // for the case of error 
      }
    if (nr) sprintf(fname, "%s%s%s_%d", argv[2], short_extension, strchr(argv[3], '_'), nr-1); 
    else strcpy(fname, argv[3]);
    in2 = fopen(fname, "rt");
    sprintf(fname, "%s%s%s", argv[2], short_extension, strchr(argv[3], '_')); 
    out = fopen(fname, "wt");
    while (TRUE) {
      if ((g = pg_readg(in2,NULL,0,&m,&n)) == NULL) break;
      cg = new col_graph(g, n);
      graph_create++;
      if (check_excess(cg, kr, kb))
        fprintf(out, "%s\n", pg_readg_line);
      else graph_stop++;
      delete cg;
      FREES(g);
    }
    fclose(out);
    fclose(in2);

    for (n = 0; n < nr_colours+2; n++)
      if (((void**)graph_dict)[n]) free_dict(((void**)graph_dict)[n]);            
    free(graph_dict);

    if (nr) 
      for (n = 0; n < nr; n++) {
        sprintf(command, "%s %s%s%s_%d", del_command, argv[2], "s", strchr(argv[3], '_'), n); 
        system(command);
      }
    fprintf(stderr, "\t %ld %ld", graph_create, graph_stop);
  }
  else if (argc == 2) { // print graphs from file
    FILE *in = fopen(argv[1], "rt");
    graph *g;
    int m, n;
    col_graph *cg;

    while (TRUE) {
      if ((g = pg_readg(in,NULL,0,&m,&n)) == NULL) break;
      cg = new col_graph(g, n);
      cg->print_graph();
      delete cg;
      FREES(g);
    }
    fclose(in);
  }
  else if (argv[1][0] == '-' && argv[1][1] == 'a') { // "amnesty" - set NON_EDGE to 0 (NON_DETERMINED)
    char fname[50];
    char buf[1000];
    graph *g;
    int m, n, nr = -1;
    col_graph *cg;
    char* cdstr;
    void *dict = NULL;

    symmetric = argc >= 4 && atol(argv[3]) == 1;
    if (symmetric) { // when r == b
      int r, b; 
      strcpy(fname, strchr(strchr(argv[2], '_')+1, '_')+1);
      sscanf(fname, "%d_%d", &r, &b);
      symmetric = r == b;
    }

    sprintf(fname, "%s.%s", argv[2], old_extension);
    sprintf(buf, "%s %s %s", mv_command, argv[2], fname);
    system(buf);

    FILE *in = fopen(fname, "rt");
    FILE *out;
    while (TRUE) {
      if ((g = pg_readg(in,NULL,0,&m,&n)) == NULL) break;
      cg = new col_graph(g, n);
      cg->col_to_zero(NON_EDGE);
      cdstr = cg->col_graph_to_graph(cg->n);
      if (symmetric) { // check complement
        char *cdstr_2 = col_graph(cg).col_graph_to_graph(cg->n);
        if (strcmp(cdstr, cdstr_2) > 0) {
          FREES(cdstr);
          cdstr = cdstr_2;
        }
        else FREES(cdstr_2);
      }
      graph_create++;
      m = mem_avail;
      inserttext(&dict, cdstr, -1);
      FREES(cdstr);
      if (m == mem_avail) graph_stop++;
      if (mem_avail < 0) {
        sprintf(fname, "%s_%d", argv[2], ++nr);
        out = fopen(fname, "wt");
        buf[0] = 0;
        print_and_free(out, dict, buf, 0, 0); // dictionary too big - dump it to a disk      
        dict = NULL;
        fclose(out);
      }
      delete cg;
      FREES(g);
    }
    sprintf(fname, "%s_%d", argv[2], ++nr);
    out = fopen(fname, "wt");
    buf[0] = 0;
    print_and_free(out, dict, buf, 0, 0); // finish - dump dictionary to a disk      
    fclose(out);
    fclose(in);
    if (nr) {
      sprintf(buf, "%s -u -m %s 0 %d -o %s", sort_command, argv[2], nr, argv[2]);
      system(buf);
      sprintf(buf, "%s %s_*", del_command, argv[2]); 
    }
    else sprintf(buf, "%s %s_0 %s", mv_command, argv[2], argv[2]);
    system(buf);

    fprintf(stderr, "\t %ld %ld", graph_create, graph_stop);
  }
  else if (argc == 3) { // create description - which edges can be selected (remove isomorphs)
    char fname[50];
    sprintf(fname, "%s%s%s", argv[1], descr_extension, strchr(argv[2], '_'));

    graph *g;
    int m, n;
    col_graph *cg;

    FILE *in = fopen(argv[2], "rt");
    FILE *out = fopen(fname, "wt");
    while (TRUE) {
      if ((g = pg_readg(in,NULL,0,&m,&n)) == NULL) break;
      cg = new col_graph(g, n);
      cg->calculate_degree();
      create_description(cg);
      fprintf(out, "%s %s\n", pg_readg_line, cg->descr);

      delete cg;
      FREES(g);
    }
    fclose(out);
    fclose(in);

    fprintf(stderr, "\t %ld %ld %ld", graph_create2, graph_create, graph_stop);
  }
  else if (argc == 5) { // create the last layer
    fname_prefix = argv[1];
    max_edges = atol(argv[2]);

    int i, j, size, esize;
    char fname[50];
    char* cdstr;
    FILE *out, *in;
    FILE *raport;
    
    sprintf(fname, "%s.txt", fname_prefix);
    raport = fopen(fname, "at"); 
    if (in = fopen(argv[3], "rt")) { 
      fscanf(in, "%d %d\n", &size, &esize);
      sprintf(fname, "%s_%d_%d_0_1", fname_prefix, max_edges, esize);
      fprintf(raport, "%d_%d_0\n", max_edges, esize);
      
      col_graph rg(size);
      for (i = 0; i < size; i++)
        for (j = 0; j < size; j++) {
          fscanf(in, "%d", &rg.g[i][j]);
          rg.g[i][j] *= 2;
        }
      
      rg.print_graph();
      
      out = fopen(fname, "wt");
      cdstr = rg.col_graph_to_graph(rg.n);
      fprintf(out, "%s\n", cdstr);
      FREES(cdstr);
      fclose(out);
    }

    if (in = fopen(argv[4], "rt")) {
      fscanf(in, "%d %d\n", &size, &esize);
      sprintf(fname, "%s_%d_0_%d_1", fname_prefix, max_edges, esize);
      fprintf(raport, "%d_0_%d\n", max_edges, esize);
      
      col_graph bg(size);
      for (i = 0; i < size; i++)
        for (j = 0; j < size; j++) {
          fscanf(in, "%d", &bg.g[i][j]);
          bg.g[i][j] *= 3;
        }
      
      bg.print_graph();
      
      out = fopen(fname, "wt");
      cdstr = bg.col_graph_to_graph(bg.n);
      fprintf(out, "%s\n", cdstr);
      FREES(cdstr);
      fclose(out);
    }
    fclose(raport);
  }
  else if (argc >= 6) { //create graphs in prevoius layer
    char fname[50];
    char command[100];
    FILE *in1, *in2;
    graph *g; 
    int m, n, i, j, the_same;
    col_graph *cg1, *cg2, *cg3;

    fname_suffix = argv[1];
    fname_prefix = argv[2];
    max_edges = atol(argv[3]);
    symmetric = argc >= 7 && atol(argv[6]) == 1;

    n = (max_edges+1)*(max_edges+1);
    graph_dict = malloc(n*sizeof(void*));
    files_count = (int *) malloc(n*sizeof(int));
    memset(graph_dict, 0, n*sizeof(void*));
    memset(files_count, 0, n*sizeof(int));

    in1 = fopen(argv[4], "rt");
    the_same = strcmp(argv[4], argv[5]) == 0; //we process the same file 
    i = 0;
    while (TRUE) {
      if ((g = pg_readg(in1,NULL,0,&m,&n)) == NULL) break;
      cg1 = new col_graph(g, n);
      FREES(g);
      cg1->calculate_degree();
      cg1->sort_by_degree();
      cg1->create_descr();

      i++;
      j = 0;
      in2 = fopen(argv[5], "rt");
      while (TRUE) {
        if ((g = pg_readg(in2,NULL,0,&m,&n)) == NULL) break;
        cg2 = new col_graph(g, n);
        FREES(g);
        cg2->calculate_degree();
        cg2->create_descr();
    
        if (cg1->n <= cg2->n) combine(cg1, cg2); 
        else {
          cg2->sort_by_degree();
          combine(cg2, cg1);
        }

        if (symmetric) {  // symmetric case - check the complement as well 
          cg3 = new col_graph(cg2);
          if (cg1->n <= cg2->n) combine(cg1, cg3);
          else combine(cg3, cg1);
          delete cg3;
        }
        delete cg2;
        j++;
        if (the_same && j > i) break; // process the same file - it is enough to do a half!
      }
      fclose(in2);
      delete cg1;
    }
    fclose(in1);
    // save results to a disk 
    save_dictionary();
    // if there were a few files pgsort will merge it into one 
    for (m = 0; m <= max_edges; m++)
      for (n = 0; n <= max_edges; n++)
        if (files_count[m*(max_edges+1)+n]) {
          sprintf(fname, "%s_%d_%d_%d_%s", fname_prefix, max_edges, m, n, fname_suffix);
          if (files_count[m*(max_edges+1)+n] > 1) {
            sprintf(command, "%s -u -m %s 0 %d -o %s", sort_command, fname, files_count[m*(max_edges+1)+n]-1, fname);
            system(command);
            sprintf(command, "%s %s_*", del_command, fname); 
          }
          else sprintf(command, "%s %s_0 %s", mv_command, fname, fname);
          system(command);
        }
    free(files_count);
    free(graph_dict);

    fprintf(stderr, "\t %ld %ld %ld %ld %ld %ld %ld %ld", est1_count, est1_stop, est2_count, est2_stop, graph_create2, graph_stop2, graph_create, graph_stop);
  }
  col_graph::free_mem();

  fprintf(stderr, " done!\n");
  return 0;
}
