///////////////////////////////////////////////////
// Ramsey On-line game
// Author: Przemek Gordinowicz 
//         Pawel Pralat 
///////////////////////////////////////////////////
//
//    Ramsey3b - (3 colour version) 
//               
//    Second program combine that merge two configurations on non_green and green edge 
//    making common edge a NON_EDGE
// 
//    The rest of comments writen in Polish, see ramsey3/readme.txt and the header of 
//    src/ramsey.cpp and src/ramrun3.cpp for more details in English
/////////////////////////////////////////////////////
//
//     Wersja 3 kolorowa!
//    1. Klasa graph zgodna z typem graph w nauty (check)
//    2. Zamiana na postać kanoniczną (fcanonise) (check)
//    3. Drzewo słownikowe, aby usuwać duplikaty (check)
//    4. Zapis wyników w kolejności alfabetycznej, później wykorzystanie "pgsort -u -m"
//    5. Odczyt i zapis z niezależnych plików z uwagi na ilości krawędzi (cz, nieb) 
//    6. Ograniczanie combine z uwagi na ilość krawędzi 
//    7. Dodany 4-ty kolor, żeby "zszywać" grafy "na raty" (najpierw czerw-nieb, poźniej ziel.)
//    8. Sortować graf G_1 (mniej razy) i wyciągnąć degree_1 na zewnątrz combine
//    9. Dodać "węgierske" sprawdzanie konfiguracji przed combine
//    10. Ograniczenie izomorficznych wywołań (descr)
//    11. Usuwanie konfiguracji podobnych do juz rozważanych (shorten)
//
//    Nazwy plików:
//    prefiks_maxkraw_ileczerw_ilenieb (np. R34_14_4_3)
//
//    Parametry (3 warianty): TODO (opis)
//    A - 1)  nazwa_pliku - wyświetla grafy (kolorowe) z pliku
//
//    B - 5)  prefiks maxkraw nczerw nnieb nziel
//    zapisuje do plików: prefiks_maxkraw_ileczerw_0_0
//                        prefiks_maxkraw_0_ile_nieb_0
//                        prefiks_maxkraw_0_0_ile_ziel
//    odpowiednie monochromatyczne kliki
//
//    C - 6) tmpnrprocesu prefiks maxkraw fname1 fname2 sym
//
//    czyta z plików: prefiks_maxkraw1_ileczerw1_ilenieb1
//                    prefiks_maxkraw2_ileczerw2_ilenieb2
//    zapisuje do plików: prefiks_maxkraw_ileczerw_ilenieb_ileziel_tmpnrprocesu
//    korzysta z tymczasowych: prefiks_maxkraw_ileczerw_ilenieb_ileziel_tmpnrprocesu_nrkolejny
//      
//    jeśli sym jest ustawione i sym = 1 to symetryczny przypadek - wybieramy graf (zamieniając kolory) aby:
//         - czerwonych krawędzi było co najmniej tyle co niebieskich
//         - jeśli jest ich tyle samo to postać kanoniczna jest pierwsza w alfabecie  
//
#include <stdio.h>
#include <stdlib.h>
#include "gtools.h"
#include "ramsey.h"

// jak zamiast 1 wstawie 0 to non_edge i not_determined to będzie to samo
// 2 oznacza "non-green"
#define NON_EDGE 1

const int nr_colours = 4;
// ilość warstw = ceil(log_2(nr_colors+2))
const int levels = 3; 
// rozmiar węzła drzewka słownikowego
const int node_size = 8 * sizeof(void *);

int max_edges, symmetric;

char *fname_prefix, *fname_suffix;
char *description, *pg_readg_line;

int max_memory = 1024*1024*768;
int mem_avail;

void *graph_dict;
int *files_count;

// Globalne, żeby nie tracić czasu na przydział pamięci
// zakładam, że wierzchołków (i krawędzi) jest nie wiecej niż 100
int used[100];
int assign[100];

#include "pg_nauty.c"

int max(int a, int b) {
  return (a > b ? a : b);
}

int min(int a, int b) {
  return (a > b ? b : a);
}

///////////////////////////////////////////////////
// coloured graph
///////////////////////////////////////////////////
class col_graph
{
public:
  int n;
  int ** g;
  int edges[nr_colours+2];
  int *(degree[nr_colours+2]);
  char *descr;
// 0 - not determined
// 1 - non-edge
// 2 - temp_almost_non-edge
// 3 - first colour
// 4 - second colour
// ...
// nr_colours+1 - last colour
  static void free_mem();
  col_graph(int size);
  col_graph(graph *_g, int size);
  col_graph(col_graph *cg, int red, int blue, int green);
  ~col_graph();
  void print_graph();
  void remove_isolated(int v);
  void switch_vertices(int u, int v);
  char* col_graph_to_graph(int size);
  void calculate_degree();
  void sort_by_degree();
  void create_descr();
};

DYNALLSTAT(graph,_g, _g_sz);    // globalne zmienne pomocnicze do col_graph_to_graph
DYNALLSTAT(graph,_h,_h_sz);      // żeby nie przydzielać w kółko pamięci
DYNALLSTAT(char,_fmt,_fmt_sz);

void col_graph::free_mem() {
  DYNFREE(_fmt, _fmt_sz);
  DYNFREE(_h, _h_sz);
  DYNFREE(_g, _g_sz);
}

// create empty graph on n vertices
col_graph::col_graph(int size) {
  int i;

  n = size;
  g = new int* [n];
  g[0] = new int [n*n];
  memset(g[0], 0, n*n * sizeof(int));
  for (i = 1 ; i < n ; i++)
    g[i] = g[i-1]+n;
  degree[0] = NULL;
  descr = NULL;
}

//tworzy graf z kolorowymi krawędziami (col_graph)
//z grafu nauty (po "size/levels" kolorowych wierzchołków w każdej z warstw)
col_graph::col_graph(graph *_g, int size) {
  n = size/levels;
  g = new int* [n];
  g[0] = new int [n*n];
  degree[0] = NULL;
  if (description) {
    descr = new char[strlen(description)+1];
    strcpy(descr, description);
  }
  else descr = NULL;

  int i, j, r, x;
  int _m;  
  int *perm = new int[n];
  set *_gv;
    
  memset(g[0], 0, n*n * sizeof(int));
  perm[0] = 0;
  for (i = 1 ; i < n ; i++) {
    g[i] = g[i-1]+n;
    perm[i] = i;
  }
  _m = (size + WORDSIZE - 1) / WORDSIZE;

  x = 1;
  for (r = 0; r < n*levels; r+=n) {
    for (i = 0; i < n; i++) {
      _gv = GRAPHROW(_g, perm[i], _m);
      for (j = i+1; j < n; j++)
        if (ISELEMENT(_gv, perm[j]))
          g[i][j] = (g[j][i] |= x);
      // pozycja wierzchołka odpowiadającego i-temu w następnej warstwie
      // sprawdzić, czy nie ma jakiejś zależności pozwalającej obliczyć to szybciej
      if (r < n*(levels-1))
      for (j = r + n; j < r + 2*n; j++)
        if (ISELEMENT(_gv, j)) {
          perm[i] = j;
          break;
        }
    }      
    x = x << 1;
  }
  delete[] perm;
}

// tworzy graf symetryczny (zamiana kolorów wg perm) - działa tylko dla trzech kolorów
col_graph::col_graph(col_graph *cg, int red, int blue, int green) {
  int i,j;
  int perm[] = {0,1,2,3,4,5};
  perm[red] = 3; 
  perm[blue] = 4;
  perm[green] = 5;

  n = cg->n;
  g = new int* [n];
  g[0] = new int [n*n];
  for (i = 1 ; i < n ; i++)
    g[i] = g[i-1]+n;

  for (i = 0 ; i < n ; i++) 
    for (j = 0 ; j < n ; j++)
      g[i][j] = perm[cg->g[i][j]];

  if (cg->degree[0]) {
    degree[0] = new int[100*(nr_colours+2)];
    for (j = 1; j < nr_colours+2; j++)
      degree[j] = degree[j-1]+100;
    for (j = 0; j < nr_colours+2; j++) {
      memcpy(degree[perm[j]], cg->degree[j], 100*sizeof(int));      
      edges[perm[j]] = cg->edges[j];
    }
  }
  else degree[0] = NULL;

  if (cg->descr) {
    descr = new char[strlen(cg->descr)+1];
    strcpy(descr, cg->descr);
  }
  else descr = NULL;
}

col_graph::~col_graph() {
  if (degree[0]) delete[] degree[0];
  if (descr) delete[] descr; 
  delete[] g[0];
  delete[] g;
}

void col_graph::print_graph() {
  int i,j;
  for (i = 0 ; i < n ; i++) {
    for (j = 0 ; j < i ; j++)
      printf("%d", g[i][j]);
    printf("\n");
  }
}

//jeśli wierzchołek v ma zerowy kolorowy stopień to go wyrzucamy
// Uwaga! wiem, że degree[0] == NULL
void col_graph::remove_isolated(int v) {
  int i, d;

  d = 0;
  for (i = 0; i < n; i++)
    if (g[i][v] > 1) d++;

  if (!d) { //wyrzucamy
    n--;
    for (i = 0 ; i <= n; i++)
      memcpy(g[i]+v, g[i]+(v+1), (n-v)*sizeof(int));
    if (!v) memcpy(g[0], g[1], n*sizeof(int));
    else g[v] = g[v+1];
    for (i = v+1 ; i < n; i++)
      g[i] = g[i+1];
  }
}

// zamienia (permutuje) wierzchołki o numerach u i v - żeby zwalnianie pamięci dobrze działało zawsze muszę wrócić
// Uwaga! wiem, że degree[0] != NULL
void col_graph::switch_vertices(int u, int v) {
  int i, x;
  int *row;

  // zamiana wierszy
  row = g[u]; g[u] = g[v]; g[v] = row;
  //zamiana kolumn
  for (i = 0; i <n; i++) {
    x = g[i][u]; g[i][u] = g[i][v]; g[i][v] = x;
  }
  for (i = 0; i < nr_colours+2; i++) {
    x = degree[i][u]; degree[i][u] = degree[i][v]; degree[i][v] = x;
  }
}

//konwertuje graf z kolorowymi krawędziami (col_graph)
//na graf nauty (o "size" kolorowych wierzchołków w każdej z warstw)
//do tego przekształca go na izomorficzną postać kanoniczną (fcanonise)
//i zwraca w postaci tekstu
char* col_graph::col_graph_to_graph(int size) {
  int _m, _n;
  int i, j, k, r;
  char *_cdstr;
  set *_gv;

  _n = n * levels;
  _m = (_n + WORDSIZE - 1) / WORDSIZE;
  
  DYNALLOC2(graph,_g,_g_sz,_m,_n,"malloc");
  DYNALLOC2(graph,_h,_h_sz,_m,_n,"malloc");
  DYNALLOC1(char, _fmt, _fmt_sz, _n+1,"malloc"); 

  //konwersja do _g, wg nug str 25;
  for (i = 0; i < _n; i++) {
    _gv = GRAPHROW(_g, i, _m);
    EMPTYSET(_gv, _m);
    // połączenie warstw
    if (i < _n - n)  ADDELEMENT(_gv, i+n);
    if (i >= n) ADDELEMENT(_gv, i-n);
  }
  for (i = 0; i < n; i++)
    for (j = 0; j < i; j++) 
      if (k = g[i][j]) {
        r = 0;
        do {
          if (k % 2) {
            _gv = GRAPHROW(_g, i + r, _m);
            ADDELEMENT(_gv, j + r);
            _gv = GRAPHROW(_g, j + r, _m);
            ADDELEMENT(_gv, i + r);
          }
          r += n;
        } while (k = k >> 1);
      }
  // kolory warstw
  if (!size) for (i = 0; i < levels; i++) 
      for (j = 0; j < n; j++)
      _fmt[i*n+j] = 2*(char)i + (used[j] ? '0' : '1');
  else for (i = 0; i < levels; i++) 
    for (j = 0; j < n; j++)
      _fmt[i*n+j] = (char)i + '0';
  
  _fmt[_n] = 0;
  // postać kanoniczna
  fcanonise(_g, _m, _n, _h, _fmt, false);
  _cdstr = pg_ntog6(_h, _m, _n);
  
  return _cdstr;
}

// oblicza stopień i ilość krawędzi w grafie
void col_graph::calculate_degree() { 
  int i, j;

  degree[0] = new int[100*(nr_colours+2)];
  memset(degree[0], 0, 100*(nr_colours+2)*sizeof(int));
  for (j = 1 ; j < nr_colours+2 ; j++)
    degree[j] = degree[j-1]+100;
  
  for (i = 0; i < n; i++)
    for (j = i+1; j < n; j++) {
      degree[g[i][j]][i]++;
      degree[g[i][j]][j]++;
    }
  memset(edges, 0, (nr_colours+2)*sizeof(int));
  for (j = 0; j < nr_colours+2; j++) {
    for (i = 0; i < n; i++)
      edges[j] += degree[j][i];
    edges[j] = edges[j] / 2;
  }
}

// sortuje wierzchołki malejąco wg stopnia - 3 kolory - NON_EDGE równe 1 (non-green równe 2 - najważniejsze)
void col_graph::sort_by_degree() { 
  int *score = new int[n];
  int *perm = new int[n];
  int i, j, k;

  for (i = 0; i < n; i++) {
    int colours[nr_colours+2] = {0, 1, 400, 2, 2, 2};
    perm[i] = i;

    if (degree[3][i] < degree[4][i]) colours[4]++; else colours[3]++;
    if (degree[3][i] < degree[5][i]) colours[5]++; else colours[3]++;
    if (degree[4][i] < degree[5][i]) colours[5]++; else colours[4]++;

    score[i] = 0;
    for (k = 0; k < nr_colours+2; k++)
      score[i] += colours[k] * degree[k][i];
  }

  //sortowanie przez wstawianie
  for (i = 1; i < n; i++) {
    k = perm[i];
    for (j = i; j > 0 && score[k] > score[perm[j-1]]; j--)
      perm[j] = perm[j-1];
    perm[j] = k;
  }
  
  //zastosowanie permutacji sortującej
  int **h = new int* [n];
  h[0] = new int [n*n];
  for (i = 1 ; i < n ; i++)
    h[i] = h[i-1]+n;
  for (i = 0 ; i < n ; i++) 
    for (j = 0 ; j < n ; j++)
      h[i][j] = g[perm[i]][perm[j]];
  delete[] g[0];
  delete[] g;
  g = h;
  
  for (j = 0; j < nr_colours+2; j++) {
    for (i = 0; i < n; i++)
      score[i] = degree[j][perm[i]];
    for (i = 0; i < n; i++)
      degree[j][i] = score[i];
  }

  if (descr) {
    char *p = descr;
    for (i = 0; i < n; i++)
      score[perm[i]] = i;
    while (*p) {
      i = (score[*p-'0']); 
      j = (score[*(p+1)-'0']); 
      if (i < j) {
        *(p++) = char(i) + '0';
        *(p++) = char(j) + '0';
      }
      else {
        *(p++) = char(j) + '0';
        *(p++) = char(i) + '0';
      }
    }
  }

  delete[] perm;
  delete[] score;
}

void col_graph::create_descr() {
  if (!descr) {
    int i, j;
    char *p;
    i = edges[3] + edges[4] + edges[5];
    descr = new char[2*i+1];
    p = descr;

    for (i = 0; i < n; i++)
      for (j = i+1; j < n; j++)
        if (g[i][j] > NON_EDGE) {
          *(p++) = '0'+ (char)i;
          *(p++) = '0'+ (char)j;
        }
    *p = 0;
  }
}

///////////////////////////////////////////////////
// drzewa słownikowe
///////////////////////////////////////////////////
// Zawierają napisy (grafy w formacie tekstowym nauty) z alfabetu 64-elementowego
// Nie jest możliwe, żeby jeden napis był prefiksem drugiego (napisy o tym samym prefiksie mają identyczną długość) 
//
// Struktura:
//       - węzeł zawiera 8 rozgałęzień (wskaźników void *) - odpowiadających 3 bitom alfabetu
//       - 1 wskaźnik (void *) wkazujący na siebie (znacznik) i tekst (końcówkę w pełnych znakach)
//       - zmienna mem_avail informuje o pamięci dostępnej na bufor;
//////////////////////////////////////////////////

// wstawia napis bit==-1 - starsza część (3bity), bit != -1 - młodsza część (3bity)
void inserttext(void *node, char *suff, char bit) {
  if (*((void **)node) == NULL) {
    size_t sz = strlen(suff) + 1 + sizeof(node);
    void *p = malloc(sz);
    mem_avail -= sz;

    *(void **)p = p;
    strcpy((char*)(((void**)p)+1), suff);
    *(void **)node = p; 
  }
  else if (*(void**)node == **(void***)node) {// ups - węzeł z tekstem
    if (strcmp(suff, (char*)((*(void***)node)+1))) { // są różne - jeśli takie same to ignorujemy
      void *old = *(void **)node;
      char *suff2 = (char*)((void**)old+1);
      mem_avail += strlen(suff2) + 1 + sizeof(node); // odzysk pamięci "na kredyt" - rzeczywiste zwolnienie nastąpi za chwilę
      char bit1, bit2;

      do {
        void *p = malloc(node_size);
        mem_avail -= node_size;
        memset(p, 0, node_size);
        *(void **)node = p; 
        
        if (bit == -1) {
          bit1 = bit = ((*suff)-63) >> 3;
          bit2 = ((*suff2)-63) >> 3;
        }
        else {
          bit = -1;
          bit1 = ((*(suff++))-63) & 7;
          bit2 = ((*(suff2++))-63) & 7;
        }        
                 
        if (bit1 == bit2) node = (void *)(*((void***)node) + bit1);
      } while (bit1 == bit2);
      if (bit == -1) {
        inserttext((void *)(*((void***)node) + bit1), suff, -1);
        inserttext((void *)(*((void***)node) + bit2), suff2, -1);
      }
      else {
        inserttext((void *)(*((void***)node) + bit1), suff, bit1);
        inserttext((void *)(*((void***)node) + bit2), suff2, bit2);
      }
      free(old);
    }
  }
  else {
    if (bit == -1) inserttext((void *)(*((void***)node) + (((*suff)-63) >> 3)), suff, ((*suff)-63) >> 3);
    else inserttext((void *)(*((void***)node) + (((*suff)-63) & 7)), suff+1, -1);
  }
}

void print_and_free(FILE *out, void *node, char* prefix, int n, char bit) {
  if ((*(void **)node) == node) { // węzeł z tekstem
    fprintf(out, "%s%s\n", prefix, (char*)(((void**)node)+1));
    mem_avail += strlen((char*)(((void**)node)+1)) + 1 + sizeof(node);
    free(node);
  }
  else {
    for (char i = 0; i < 8; i++)
      if (*((void**)node+i)) 
        if (n & 1) {
          prefix[n/2+1] = 0;
          prefix[n/2] = 63 + ((bit << 3) | i);
          print_and_free(out, *((void**)node+i), prefix, n+1, 0);
          prefix[n/2] = 0;
        }
        else {
          print_and_free(out, *((void**)node+i), prefix, n+1, i);
        }
    prefix[n] = 0;
    mem_avail += node_size;
    free(node);
  }
  return;
}

void free_dict(void *node) {
  if ((*(void **)node) == node) { // węzeł z tekstem
    mem_avail += strlen((char*)(((void**)node)+1)) + 1 + sizeof(node);
    free(node);
  }
  else {
    for (char i = 0; i < 8; i++)
      if (*((void**)node+i)) 
        free_dict(*((void**)node+i));
    mem_avail += node_size;
    free(node);
  }
  return;
}

// zapis do plików zawartości drzew słownikowych (graph_dict...)
void save_dictionary(void) {
  int i, j, k;
  char fname[100];
  char buf[1000];
  FILE *out;

  for (i = 0; i <= max_edges; i++)
    for (j = 0; j <= max_edges; j++) 
      for (k = 0; k <= max_edges; k++) 
        if (((void **)graph_dict)[(i*(max_edges+1)+j)*(max_edges+1)+k]) {
          sprintf(fname, "%s %d_%d_%d_%d >>%s.txt", echo_command, max_edges, i, j, k, fname_prefix);
          system(fname);
          sprintf(fname, "%s_%d_%d_%d_%d_%s_%d", fname_prefix, max_edges, i, j, k, fname_suffix, files_count[(i*(max_edges+1)+j)*(max_edges+1)+k]++);
          out = fopen(fname, "wt");
          buf[0] = 0;
          print_and_free(out, ((void **)graph_dict)[(i*(max_edges+1)+j)*(max_edges+1)+k], buf, 0, 0);        
          ((void **)graph_dict)[(i*(max_edges+1)+j)*(max_edges+1)+k] = NULL;
          fclose(out);
        }
}

///////////////////////////////////////////////////
// Hungarian algorithm to better estimate bound for number of edges
///////////////////////////////////////////////////
int min_element(int *r, int n) {
  int x = r[0], i;
  for (i = 1; i < n; i++)
    if (x > r[i]) x = r[i];
  return x;
}

int* hungarian(int **C, int n) {
  // finds the cheapest matching, with respect to C as a matrix of costs
  int *matchx = new int[n], *matchy = new int[n];
  int *corx = new int[n], *cory = new int[n];
  int *markx = new int[n], *marky = new int[n];
  int *minval = new int[n], *choice = new int[n];
  int *queue = new int[n];
  int i, j, k; 

  for (i = 0; i < n; i++) {
    matchx[i] = matchy[i] = -1;
    cory[i] = 0; 
    corx[i] = min_element(C[i], n);
  }
  for (k = 0; k < n; k++) {
    int z = - 1, qb = 0, qe = 0;
    for (i = 0; i < n; i++) {
      markx[i] = marky[i] = 0;
      minval[i] = -1;
      if (matchx[i] == -1) queue[qe++] = i; 
    }
    while (z == -1) {  // no vertex for correction 
      while (qb < qe) {  // queue is not empty
        i = queue[qb++];
        markx[i] = 1;
        for (j = 0; j < n; j++) { // looking for matching for x_i
          if (!marky[j] && (minval[j] == -1 || minval[j] > C[i][j] - corx[i] - cory[j])) {
            choice[j] = i;
            minval[j] = C[i][j] - corx[i] - cory[j];
            if (minval[j] == 0) { // we have matching
              marky[j] = 1;
              if (matchy[j] != -1) queue[qe++] = matchy[j]; //ups! y_j is already matched
              else {
                z = j;
                break;
              }
            }
          }
        }
        if (z != -1) break;
      }
      if (z == -1) { // cannot match more - substract minimal uncovered (marked) value
        int v = -1;
        for (i = 0; i < n; i++)
          if (!marky[i] && (v == -1 || minval[i] < v)) v = minval[i];
        for (i = 0; i < n; i++) {
          if (markx[i]) corx[i] += v; //test
          if (marky[i]) cory[i] -= v;
          else if ((minval[i] -= v) == 0) {
            marky[i] = 1;
            if (matchy[i] != -1) queue[qe++] = matchy[i]; // another matching to check (undo)
            else z = i; // yeah, we have new mathching
          }
        }
      }
    }
    while (z != -1) {
      int y = matchx[choice[z]];
      matchx[choice[z]] = z;
      matchy[z] = choice[z];
      z = y;
    }
  }

  delete[] queue;
  delete[] minval; delete[] choice;
  delete[] markx; delete[] marky;
  delete[] cory; delete[] corx; 
  delete[] matchy;

  return matchx;
}

///////////////////////////////////////////////////
// the algorithm
///////////////////////////////////////////////////
int edges3[nr_colours+2];   // ilość krawędzi "w zszytej części" - zmienna globalna ma być wypełnione zerami

int est_edges(col_graph * g1, col_graph * g2, int v) { // szacuje ilość krawędzi - wierzchołki 0..v z g1 - assigned
  int r, j;
  int est = 0;

  for (r = NON_EDGE+2; r < nr_colours+2; r++) {
    for (j = 0; j <= v; j++) 
      est += max(g1->degree[r][j], g2->degree[r][assign[j]]);
    est += max(g1->edges[r], g2->edges[r]);
    est += edges3[r];
  }
  return est;
}

int estimate_edges(col_graph * g1, col_graph * g2, int v) { // szacuje ilość krawędzi metodą węgierską - wierzchołki 0..v-1 z g1 - assigned
  int r, i, j, k;
  int est = 0;
  int size = g2->n;

  for (j = 0; j < v; j++)      // size - większa z ilości nieprzypisanych wierzchołków z g1 i g2
    if (assign[j] < g2->n) size--;
  if (size < g1->n - v) size = g1->n - v; 

  int **C = new int *[size], **x = new int *[size], **y  = new int *[size];
  C[0] = new int[size*size]; 
  x[0] = new int[(v+1) * nr_colours * size];
  y[0] = new int[(v+1) * nr_colours * size];
  memset(x[0], 0, size * (v+1) * nr_colours * sizeof(int));
  memset(y[0], 0, size * (v+1) * nr_colours * sizeof(int));
  memset(C[0], 0, size * size * sizeof(int));

  for (i = 1; i < size; i++) {
    C[i] = C[i-1] + size;
    x[i] = x[i-1] + ((v+1) * nr_colours);
    y[i] = y[i-1] + ((v+1) * nr_colours);
  }

  // x - stan w grafie g2
  for (i = 0, k = 0; i < g2->n; i++)
    if (!used[i]) {
      for (r = NON_EDGE+2; r < nr_colours+2; r++)
        x[k][(v) * nr_colours + r - NON_EDGE-1] = g2->degree[r][i];
      for (j = 0; j < v; j++)
        if (assign[j] < g2->n && g2->g[i][assign[j]] > NON_EDGE) {
          x[k][j * nr_colours + g2->g[i][assign[j]] - NON_EDGE-1] = 1;
          x[k][(v) * nr_colours + g2->g[i][assign[j]] - NON_EDGE-1]--;
        }
      k++;
    }
  // y - stan w grafie g1
  for (i = v; i < g1->n; i++) {
    for (r = NON_EDGE+2; r < nr_colours+2; r++)
      y[i-v][(v) * nr_colours + r - NON_EDGE-1] = g1->degree[r][i];
      for (j = 0; j < v; j++)
        if (g1->g[i][j] > NON_EDGE) {
          y[i-v][j * nr_colours + g1->g[i][j] - NON_EDGE-1] = 1;
          y[i-v][(v) * nr_colours + g1->g[i][j] - NON_EDGE-1]--;
        }
  }

  //macierz C
  for (i = 0; i < size; i++)
    for (j = 0; j < size; j++)
      for (r = 0; r < (v+1)*nr_colours; r++)
        C[i][j] += abs(x[i][r] - y[j][r]);
  int *matchx = hungarian(C, size); 

  //wartość po kolorach
  for (i = 0; i < size; i++) 
    for (r = 0; r < (v+1)*nr_colours; r++) {
      est += max(x[i][r], y[matchx[i]][r]);
  }
  delete[] matchx; 

  // tutaj to co liczymy klasycznie
  for (r = NON_EDGE+2; r < nr_colours+2; r++) {
    for (j = 0; j < v; j++) 
      est += max(g1->degree[r][j], g2->degree[r][assign[j]]);
    est += 2 * edges3[r];
  }

  delete[] y[0];
  delete[] x[0];
  delete[] C[0];
  delete[] y;
  delete[] x;
  delete[] C;

  return (est+1)/2;
}

void combine(col_graph *g1, col_graph *g2, int w1, int w2, int v1, int v2) {
  // v_i is associated with w_i, i=1,2
  int i, j, r, v, size;
  int still_good; 
  int limit = 3*(g1->n)/5 + 1; // nr wierzchołka od którego nie jest sprawdzana ilość krawędzi metodą węgierską

  memset(used, 0, g2->n * sizeof(int));
  used[v1] = used[v2] = 1;

  memset(assign, -1, g1->n * sizeof(int));     // jak się wypełni -1 (byte) to będą też -1 int
  assign[w1] = v1;
  assign[w2] = v2;

  if (estimate_edges(g1, g2, 2) > max_edges) return; 

  v = 2;
  size = g2->n;
  do {
    if (v == g1->n) { 
      col_graph ng(size);
      for (i = 0; i < g2->n; i++)
        memcpy(ng.g[i], g2->g[i], g2->n * sizeof(int));

      for (i = 0; i < v; i++) 
        for (j = i+1; j < v; j++)
          if (g1->g[i][j])
            ng.g[assign[i]][assign[j]] = ng.g[assign[j]][assign[i]] = g1->g[i][j];
      ng.g[v1][v2] = ng.g[v2][v1] = 0; //tmp NON_EDGE;  

      // count the number of edges
      int edges[nr_colours+2];
      memcpy(edges, edges3, (nr_colours+2)*sizeof(int));
      for (r = NON_EDGE+2; r < nr_colours+2; r++) {
        for (j = 0; j < v; j++) 
          edges[r] += g2->degree[r][assign[j]] + g1->degree[r][j]; // krawędzie między związaną a niezwiązaną częścią
        edges[r] += g2->edges[r];                                   // krawędzie w niezwiązanej części 
      }

      ng.remove_isolated(v2);
      ng.remove_isolated(v1);

      char *cdstr = ng.col_graph_to_graph(size);
      // jeśli problem symetryczny to krawędzi czerwonych nie mniej niż niebieskich \le zielonych
      // a jeśli tyle samo, to napis (kanoniczny) nie większy (w sensie alfabetycznym)
      int red=3, blue=4, green=5;
      if (symmetric) {
        i=0;
        //najpierw po krawędziach
        if (edges[red] < edges[blue]) {
          i = red; red = blue; blue = i;
        }
        if (edges[blue] < edges[green]) {
          i = blue; blue = green; green = i;
          if (edges[red] < edges[blue]) {
            i = red; red = blue; blue = i;
          }
        }
        if (i) { // była przynajmniej jedna zamiana!
          char *cdstr_2 = col_graph(&ng, red, blue, green).col_graph_to_graph(size);
          FREES(cdstr);
          cdstr = cdstr_2;
        }
        if (edges[red] == edges[blue]) {
          if (edges[blue] == edges[green]) { //tu chyba trzeba srawdzić wszystkie permutacje 
            char *cdstr_2 = col_graph(&ng, red, green, blue).col_graph_to_graph(size);
            if (strcmp(cdstr, cdstr_2) > 0) {
              FREES(cdstr);
              cdstr = cdstr_2;
            }
            else FREES(cdstr_2);
            cdstr_2 = col_graph(&ng, blue, green, red).col_graph_to_graph(size);
            if (strcmp(cdstr, cdstr_2) > 0) {
              FREES(cdstr);
              cdstr = cdstr_2;
            }
            else FREES(cdstr_2);
            cdstr_2 = col_graph(&ng, green, red, blue).col_graph_to_graph(size);
            if (strcmp(cdstr, cdstr_2) > 0) {
              FREES(cdstr);
              cdstr = cdstr_2;
            }
            else FREES(cdstr_2);
            cdstr_2 = col_graph(&ng, green, blue, red).col_graph_to_graph(size);
            if (strcmp(cdstr, cdstr_2) > 0) {
              FREES(cdstr);
              cdstr = cdstr_2;
            }
            else FREES(cdstr_2);
          }
          // tu tylko jedną - tą ostatnią
          char *cdstr_2 = col_graph(&ng, blue, red, green).col_graph_to_graph(size);
          if (strcmp(cdstr, cdstr_2) > 0) {
            FREES(cdstr);
            cdstr = cdstr_2;
          }
          else FREES(cdstr_2);
        }
        else if (edges[blue] == edges[green]) {
          // tu też tylko jedną
          char *cdstr_2 = col_graph(&ng, red, green, blue).col_graph_to_graph(size);
          if (strcmp(cdstr, cdstr_2) > 0) {
            FREES(cdstr);
            cdstr = cdstr_2;
          }
          else FREES(cdstr_2);
        }
      }
      inserttext((void**)graph_dict+((edges[red]*(max_edges+1)+edges[blue])*(max_edges+1)+edges[green]), cdstr, -1);
      FREES(cdstr);
      if (mem_avail < 0) save_dictionary(); // słownik zajmuje za dużo - zrzut na dysk      
      v--;
    }
    if (assign[v] < g2->n) {  // wierzchołek v ostatnio miał przypisany niski numer lub nie był przypisany - iść dalej
      still_good = 0;
      if (assign[v] >= 0) {    // zwalniamy wierzchołek w g2
        i = assign[v];
        used[i] = 0; 
        //zliczanie krawędzi
        for (r = 0; r < nr_colours+2; r++) {
          g1->edges[r] += g1->degree[r][v];
          g2->edges[r] += g2->degree[r][i];
        }
        for (j = 0; j < v; j++) {
          if (assign[j] < g2->n) {
            g2->degree[g2->g[i][assign[j]]][i]++;
            g2->degree[g2->g[i][assign[j]]][assign[j]]++;
            if (g1->g[v][j] == 0) edges3[g2->g[i][assign[j]]]--;
          }
          g1->degree[g1->g[v][j]][v]++;
          g1->degree[g1->g[v][j]][j]++;
          edges3[g1->g[v][j]]--;
        }
      }
      for (i = assign[v]+1 ; i < g2->n ; i++)
        if (!used[i]) {
          still_good = 1;
          for (j = 0 ; (j < v) && still_good ; j++)
            if (assign[j] < g2->n)
              if ((g2->g[i][assign[j]] != 0) && (g1->g[v][j] != 0) && (g2->g[i][assign[j]] != g1->g[v][j]))
                still_good = 0;

          if (still_good) {
            used[i] = 1;
            assign[v] = i;
            //zliczanie krawędzi
            for (j = 0; j < v; j++) {
              if (assign[j] < g2->n) {
                g2->degree[g2->g[i][assign[j]]][i]--;
                g2->degree[g2->g[i][assign[j]]][assign[j]]--;
                if (g1->g[v][j] == 0) edges3[g2->g[i][assign[j]]]++;
              }
              g1->degree[g1->g[v][j]][v]--;
              g1->degree[g1->g[v][j]][j]--;
              edges3[g1->g[v][j]]++;
            }
            for (r = 0; r < nr_colours+2; r++) {
              g1->edges[r] -= g1->degree[r][v];
              g2->edges[r] -= g2->degree[r][i];
            }
            if (est_edges(g1, g2, v) <= max_edges)
              if (v >= limit || estimate_edges(g1, g2, v+1) <= max_edges) v++; 
            break;
          }
        }
      if (!still_good) { // nie nastąpiło przypisanie wierzchołka - assign a new vertex
        assign[v] = size++;
        //zliczanie krawędzi
        for (j = 0; j < v; j++) {
          g1->degree[g1->g[v][j]][v]--;
          g1->degree[g1->g[v][j]][j]--;
          edges3[g1->g[v][j]]++;
        }
        for (r = 0; r < nr_colours+2; r++) 
          g1->edges[r] -= g1->degree[r][v];
        if (est_edges(g1, g2, v) <= max_edges)
          if (v >= limit || estimate_edges(g1, g2, v+1) <= max_edges) v++; 
      }
    }
    else {  // wierzchołek v miał wysoki numer - poza g1 - wycofać v
      //zliczanie krawędzi
      for (r = 0; r < nr_colours+2; r++) 
        g1->edges[r] += g1->degree[r][v];
      for (j = 0; j < v; j++) {
        g1->degree[g1->g[v][j]][v]++;
        g1->degree[g1->g[v][j]][j]++;
        edges3[g1->g[v][j]]--;
      }
      size--;
      assign[v] = -1;
      v--;
    }  
  } while (v > 1);
}

void combine(col_graph * g1, col_graph * g2) {
  int k, m, r, s, s00, s01, s10, s11;
  char *p2;

  g1->degree[2][0]--;
  g1->degree[2][1]--;
  g1->edges[2]--;        
  for (r = 0; r < nr_colours+2; r++) 
    g1->edges[r] -= g1->degree[r][0] + g1->degree[r][1];

  p2 = g2->descr;
  while (*p2) {
    k = *(p2++) - '0';
    m = *(p2++) - '0';
    if (g2->g[k][m] == 5) { // skleca niezielony (2 = prawie non-edge) z zielonym = 5!
      g2->degree[5][k]--;
      g2->degree[5][m]--;
      g2->edges[5]--;
      for (r = 0; r < nr_colours+2; r++) 
        g2->edges[r] -= g2->degree[r][k] + g2->degree[r][m];

      s = s00 = s01 = s10 = s11 = 0;        // zliczamy krawędzie, żeby sprawdzić, czy jest szansa  
      for (r = NON_EDGE+2; r < nr_colours+2; r++) {  // na "zszycie" odpowiednio małego grafu
        s += max(g1->edges[r], g2->edges[r]);
        s00 += max(g1->degree[r][0], g2->degree[r][k]);
        s11 += max(g1->degree[r][1], g2->degree[r][m]);
        s01 += max(g1->degree[r][0], g2->degree[r][m]);
        s10 += max(g1->degree[r][1], g2->degree[r][k]);
      }
      if (g1->n > g2->n) {
        if (k != 0)  g2->switch_vertices(0, k);    // permutacja wierzchołków, w której k,m są na początku
        if (m != 1)  g2->switch_vertices(1, m);    // robiona przez dwie zamiany (0<->k i 1<->m)

        if (s+s00+s11 <= max_edges) combine(g2,g1,0,1,0,1);  
        if (s+s01+s10 <= max_edges) combine(g2,g1,1,0,0,1); 

        if (m != 1) g2->switch_vertices(1, m);    // powrót do identyczności 
        if (k != 0)  g2->switch_vertices(0, k);    // zamiany w odwrotnej kolejności (1<->m i 0<->k) 
      }
      else {
        if (s+s00+s11 <= max_edges) combine(g1,g2,0,1,k,m);  
        if (s+s01+s10 <= max_edges) combine(g1,g2,1,0,k,m); 
      }
      for (r = 0; r < nr_colours+2; r++) 
        g2->edges[r] += g2->degree[r][k] + g2->degree[r][m];
      g2->degree[5][k]++;
      g2->degree[5][m]++;
      g2->edges[5]++;
    }
  }
  for (r = 0; r < nr_colours+2; r++) 
    g1->edges[r] += g1->degree[r][0] + g1->degree[r][1];
  g1->degree[2][0]++;
  g1->degree[2][1]++;
  g1->edges[2]++;
}

///////////////////////////////////////////////////
// Main program
///////////////////////////////////////////////////
int main(int argc, char * argv[])
{
  for (int i = 1; i < argc; i++)
    fprintf(stderr, "%s ", argv[i]);
  mem_avail = max_memory;
  if (argc == 2) { // print graphs from file
    FILE *in = fopen(argv[1], "rt");
    graph *g;
    int m, n;
    col_graph *cg;

    while (TRUE) {
      if ((g = pg_readg(in,NULL,0,&m,&n)) == NULL) break;
      cg = new col_graph(g, n);
      cg->print_graph();
      delete cg;
      FREES(g);
    }
    fclose(in);
  }
  if (argc == 5) { // permute colors (3,4,5 == r,b,g)
    FILE *in = fopen(argv[1], "rt");
    graph *g;
    int m, n;
    col_graph *cg;

    while (TRUE) {
      if ((g = pg_readg(in,NULL,0,&m,&n)) == NULL) break;
      cg = new col_graph(g, n);
      printf("%s\n", col_graph(cg, atol(argv[2]), atol(argv[3]), atol(argv[4])).col_graph_to_graph(cg->n));
      delete cg;
      FREES(g);
    }
    fclose(in);
  }
  if (argc == 6) { // create the last layer - dla symetrycznych w drugiej i trzeciej warstwie 0
    fname_prefix = argv[1];
    max_edges = atol(argv[2]);

    int i, j, size;
    char fname[50];
    char* cdstr;
    FILE *out;
    FILE *raport;
    
    sprintf(fname, "%s.txt", fname_prefix);
    raport = fopen(fname, "at");
    if (size = atol(argv[3])) {
      sprintf(fname, "%s_%d_%d_0_0_1", fname_prefix, max_edges, size*(size-1)/2);
      fprintf(raport, "%d_%d_0_0\n", max_edges, size*(size-1)/2);
      
      col_graph rg(size);
      for (i = 0; i < size; i++)
        for (j = i+1; j < size; j++)
          rg.g[i][j] = rg.g[j][i] = 3;
      
      rg.print_graph();
      
      out = fopen(fname, "wt");
      cdstr = rg.col_graph_to_graph(rg.n);
      fprintf(out, "%s\n", cdstr);
      FREES(cdstr);
      fclose(out);
    }

    if (size = atol(argv[4])) {
      sprintf(fname, "%s_%d_0_%d_0_1", fname_prefix, max_edges, size*(size-1)/2);
      fprintf(raport, "%d_0_%d_0\n", max_edges, size*(size-1)/2);
      
      col_graph bg(size);
      for (i = 0; i < size; i++)
        for (j = i+1; j < size; j++)
          bg.g[i][j] = bg.g[j][i] = 4;
      
      bg.print_graph();
      
      out = fopen(fname, "wt");
      cdstr = bg.col_graph_to_graph(bg.n);
      fprintf(out, "%s\n", cdstr);
      FREES(cdstr);
      fclose(out);
    }

    if (size = atol(argv[5])) {
      sprintf(fname, "%s_%d_0_0_%d_1", fname_prefix, max_edges, size*(size-1)/2);
      fprintf(raport, "%d_0_0_%d\n", max_edges, size*(size-1)/2);
      
      col_graph gg(size);
      for (i = 0; i < size; i++)
        for (j = i+1; j < size; j++)
          gg.g[i][j] = gg.g[j][i] = 5;
      
      gg.print_graph();
      
      out = fopen(fname, "wt");
      cdstr = gg.col_graph_to_graph(gg.n);
      fprintf(out, "%s\n", cdstr);
      FREES(cdstr);
      fclose(out);
    }
    fclose(raport);
  }
  else if (argc > 6) { //create graphs in prevoius layer
    char fname[50];
    char command[100];
    FILE *in1, *in2;
    graph *gr; 
    int m, n, p, i, j;
    col_graph *cg1, *cg2, *cg3;
    int r, b, g;
    int _r, _b, _g;
    int  good[6];

    fname_suffix = argv[1];
    fname_prefix = argv[2];
    max_edges = atol(argv[3]);
    symmetric = argc >= 7 && atol(argv[6]) == 1;

    n = (max_edges+1)*(max_edges+1)*(max_edges+1);
    graph_dict = malloc(n*sizeof(void*));
    files_count = (int *) malloc(n*sizeof(int));
    memset(graph_dict, 0, n*sizeof(void*));
    memset(files_count, 0, n*sizeof(int));

    // warunki CHECK - przy split w ramrun chyba OK
    strcpy(fname, strchr(strchr(argv[4], '_')+1, '_')+1);
    sscanf(fname, "%d_%d_%d", &_r, &_b, &_g);
    strcpy(fname, strchr(strchr(argv[5], '_')+1, '_')+1);
    sscanf(fname, "%d_%d_%d", &r, &b, &g);

    good[0] = g > 0 && max(r, _r) + max(b, _b) + max(g-1, _g) <= max_edges;    //r,b,g
    if (symmetric) {
      good[1] = (b > 0 && max(r, _r) + max(g, _b) + max(b-1, _g) <= max_edges);   // r,g,b
      good[2] = (g > 0 && max(b, _r) + max(r, _b) + max(g-1, _g) <= max_edges);   // b,r,g
      good[3] = (r > 0 && max(b, _r) + max(g, _b) + max(r-1, _g) <= max_edges);   // b,g,r
      good[4] = (b > 0 && max(g, _r) + max(r, _b) + max(b-1, _g) <= max_edges);   // g,r,b
      good[5] = (r > 0 && max(g, _r) + max(b, _b) + max(r-1, _g) <= max_edges);    // g,b,r
    }

    in1 = fopen(argv[4], "rt");
    i = 0;
    while (TRUE) {
      if ((gr = pg_readg(in1,NULL,0,&m,&n)) == NULL) break;
      cg1 = new col_graph(gr, n);
      FREES(gr);
      cg1->calculate_degree();
      cg1->sort_by_degree();

      i++;
      j = 0;
      in2 = fopen(argv[5], "rt");
      while (TRUE) {
        if ((gr = pg_readg(in2,NULL,0,&m,&n)) == NULL) break;
        cg2 = new col_graph(gr, n);
        FREES(gr);
        cg2->calculate_degree();
        cg2->create_descr();
        if (cg1->n > cg2->n) cg2->sort_by_degree();

        if (good[0]) combine(cg1, cg2); 

        if (symmetric) {  // przypadek symetryczny - sprawdzamy także permutacje kolorów
          if (good[1]) {
            cg3 = new col_graph(cg2, 3, 5, 4);
            combine(cg1, cg3);
            delete cg3;
          }
          if (good[2]) {
            cg3 = new col_graph(cg2, 4, 3, 5);
            combine(cg1, cg3);
            delete cg3;
          }
          if (good[3]) {
            cg3 = new col_graph(cg2, 4, 5, 3);
            combine(cg1, cg3);
            delete cg3;
          }
          if (good[4]) {
            cg3 = new col_graph(cg2, 5, 3, 4);
            combine(cg1, cg3);
            delete cg3;
          }
          if (good[5]) {
            cg3 = new col_graph(cg2, 5, 4, 3);
            combine(cg1, cg3);
            delete cg3;
          }
        }
        delete cg2;
        j++;
      }
      fclose(in2);
      delete cg1;
    }
    fclose(in1);
    // zapisać wyniki na dysk
    save_dictionary();
    // jeśli było kilka plików pgsort połączy je w jeden
    for (m = 0; m <= max_edges; m++)
      for (n = 0; n <= max_edges; n++)
        for (p = 0; p <= max_edges; p++)
        if (files_count[(m*(max_edges+1)+n)*(max_edges+1)+p]) {
          sprintf(fname, "%s_%d_%d_%d_%d_%s", fname_prefix, max_edges, m, n, p, fname_suffix);
          if (files_count[(m*(max_edges+1)+n)*(max_edges+1)+p] > 1) {
            sprintf(command, "%s -u -m %s 0 %d -o %s", sort_command, fname, files_count[(m*(max_edges+1)+n)*(max_edges+1)+p]-1, fname);
            system(command);
            sprintf(command, "%s %s_*", del_command, fname); 
          }
          else sprintf(command, "%s %s_0 %s", mv_command, fname, fname);
          system(command);
        }
    free(files_count);
    free(graph_dict);
  }
  col_graph::free_mem();

  fprintf(stderr, "done!\n");
  return 0;
}
