///////////////////////////////////////////////////
// Ramsey On-line game
// Author: Przemek Gordinowicz 
//         Pawel Pralat 
///////////////////////////////////////////////////
//
//  Part of the main file (ramsey.cpp) including the main algorithm
//
///////////////////////////////////////////////////
//
///////////////////////////////////////////////////
// Hungarian algorithm to better estimate bound for number of edges
///////////////////////////////////////////////////
int min_element(int *r, int n) {
  int x = r[0], i;
  for (i = 1; i < n; i++)
    if (x > r[i]) x = r[i];
  return x;
}

int* hungarian(int **C, int n) {
  // finds the cheapest matching, with respect to C as a matrix of costs
  int *matchx = new int[n], *matchy = new int[n];
  int *corx = new int[n], *cory = new int[n];
  int *markx = new int[n], *marky = new int[n];
  int *minval = new int[n], *choice = new int[n];
  int *queue = new int[n];
  int i, j, k; 

  for (i = 0; i < n; i++) {
    matchx[i] = matchy[i] = -1;
    cory[i] = 0; 
    corx[i] = min_element(C[i], n);
  }
  for (k = 0; k < n; k++) {
    int z = - 1, qb = 0, qe = 0;
    for (i = 0; i < n; i++) {
      markx[i] = marky[i] = 0;
      minval[i] = -1;
      if (matchx[i] == -1) queue[qe++] = i; 
    }
    while (z == -1) {  // no vertex for correction 
      while (qb < qe) {  // queue is not empty
        i = queue[qb++];
        markx[i] = 1;
        for (j = 0; j < n; j++) { // looking for matching for x_i
          if (!marky[j] && (minval[j] == -1 || minval[j] > C[i][j] - corx[i] - cory[j])) {
            choice[j] = i;
            minval[j] = C[i][j] - corx[i] - cory[j];
            if (minval[j] == 0) { // we have matching
              marky[j] = 1;
              if (matchy[j] != -1) queue[qe++] = matchy[j]; //ups! y_j is already matched
              else {
                z = j;
                break;
              }
            }
          }
        }
        if (z != -1) break;
      }
      if (z == -1) { // cannot match more - substract minimal uncovered (marked) value
        int v = -1;
        for (i = 0; i < n; i++)
          if (!marky[i] && (v == -1 || minval[i] < v)) v = minval[i];
        for (i = 0; i < n; i++) {
          if (markx[i]) corx[i] += v; //test
          if (marky[i]) cory[i] -= v;
          else if ((minval[i] -= v) == 0) {
            marky[i] = 1;
            if (matchy[i] != -1) queue[qe++] = matchy[i]; // another matching to check (undo)
            else z = i; // yeah, we have new mathching
          }
        }
      }
    }
    while (z != -1) {
      int y = matchx[choice[z]];
      matchx[choice[z]] = z;
      matchy[z] = choice[z];
      z = y;
    }
  }

  delete[] queue;
  delete[] minval; delete[] choice;
  delete[] markx; delete[] marky;
  delete[] cory; delete[] corx; 
  delete[] matchy;

  return matchx;
}

///////////////////////////////////////////////////
// the algorithm
///////////////////////////////////////////////////
int edges3[nr_colours+2];   // number of edges in a merged part - global variable (filled with 0s)

int est_edges(col_graph * g1, col_graph * g2, int v) { // estimane number of edges - vertices 0..v from g1 - assigned
  int r, j;
  int est = 0;

  est1_count++;
  for (r = NON_EDGE+1; r < nr_colours+2; r++) {
    for (j = 0; j <= v; j++) 
      est += max(g1->degree[r][j], g2->degree[r][assign[j]]);
    est += max(g1->edges[r], g2->edges[r]);
    est += edges3[r];
  }
  return est;
}

int estimate_edges(col_graph *g1, col_graph *g2, int v) { // estimate number of edges with hungarian algorithm - vertices 0..v-1 from g1 - assigned
  int r, i, j, k;
  int est = 0;
  int size = g2->n;

  est2_count++;
  for (j = 0; j < v; j++)      // size - bigger number of non assigned vertices from g1 or g2
    if (assign[j] < g2->n) size--;
  if (size < g1->n - v) size = g1->n - v; 

  int **C = new int *[size], **x = new int *[size], **y  = new int *[size];
  C[0] = new int[size*size]; 
  x[0] = new int[(v+1) * nr_colours * size];
  y[0] = new int[(v+1) * nr_colours * size];
  memset(x[0], 0, size * (v+1) * nr_colours * sizeof(int));
  memset(y[0], 0, size * (v+1) * nr_colours * sizeof(int));
  memset(C[0], 0, size * size * sizeof(int));

  for (i = 1; i < size; i++) {
    C[i] = C[i-1] + size;
    x[i] = x[i-1] + ((v+1) * nr_colours);
    y[i] = y[i-1] + ((v+1) * nr_colours);
  }

  // x - quantity in graph g2
  for (i = 0, k = 0; i < g2->n; i++)
    if (!used[i]) {
      for (r = 2; r < nr_colours+2; r++)
        x[k][(v) * nr_colours + r - 2] = g2->degree[r][i];
      for (j = 0; j < v; j++)
        if (assign[j] < g2->n && g2->g[i][assign[j]] >= 2) {
          x[k][j * nr_colours + g2->g[i][assign[j]] - 2] = 2; 
          x[k][(v) * nr_colours + g2->g[i][assign[j]] - 2]--;
        }
      k++;
    }
  // y - quantity in graph g1
  for (i = v; i < g1->n; i++) {
    for (r = 2; r < nr_colours+2; r++)
      y[i-v][(v) * nr_colours + r - 2] = g1->degree[r][i];
      for (j = 0; j < v; j++)
        if (g1->g[i][j] >= 2) {
          y[i-v][j * nr_colours + g1->g[i][j] - 2] = 2; 
          y[i-v][(v) * nr_colours + g1->g[i][j] - 2]--;
        }
  }

  //matrix C
  for (i = 0; i < size; i++)
    for (j = 0; j < size; j++)
      for (r = 0; r < (v+1)*nr_colours; r++)
        C[i][j] += abs(x[i][r] - y[j][r]);
  int *matchx = hungarian(C, size); 

  //value for each colour
  for (i = 0; i < size; i++) 
    for (r = 0; r < (v+1)*nr_colours; r++) {
      est += max(x[i][r], y[matchx[i]][r]);
  }
  delete[] matchx; 

  // here is the rest (counted simply) 
  for (r = 2; r < nr_colours+2; r++) {
    est += 2 * edges3[r];
  }

  delete[] y[0];
  delete[] x[0];
  delete[] C[0];
  delete[] y;
  delete[] x;
  delete[] C;

  return (est+1)/2;
}

void combine(col_graph *g1, col_graph *g2, int w1, int w2, int v1, int v2) {
  // creating new configuration from g1 and g2
  // v_i is associated with w_i, i=1,2
  // attempts to merging vertices to decrease numbers of coloured edges 
  int i, j, r, v, size;
  int still_good;
  int limit = 3*(g1->n)/5 + 1; // no of vertex from which number of edges is not estimated with hungarian method (formula given by trial)

  memset(used, 0, g2->n * sizeof(int));
  used[v1] = used[v2] = 1;

  memset(assign, -1, g1->n * sizeof(int));     // when filled with -1 (byte) contains as well -1 (int)
  assign[w1] = v1;
  assign[w2] = v2;

  if (estimate_edges(g1, g2, 2) > max_edges) { 
    est2_stop++;
    return; 
  }

  v = 2;
  size = g2->n;
  do {
    if (v == g1->n) {
      col_graph ng(size);
      for (i = 0; i < g2->n; i++)
        memcpy(ng.g[i], g2->g[i], g2->n * sizeof(int));

      for (i = 0; i < v; i++)
        for (j = i+1; j < v; j++)
          if (g1->g[i][j])
            ng.g[assign[i]][assign[j]] = ng.g[assign[j]][assign[i]] = g1->g[i][j];    
      ng.g[v1][v2] = ng.g[v2][v1] = NON_EDGE;  

      // count the number of edges
      int edges[nr_colours+2];
      memcpy(edges, edges3, (nr_colours+2)*sizeof(int));           // edges in merged part 
      for (r = NON_EDGE+1; r < nr_colours+2; r++) {
        for (j = 0; j < v; j++) 
          edges[r] += g2->degree[r][assign[j]] + g1->degree[r][j]; // edges between merged and non merged part 
        edges[r] += g2->edges[r];                                  // edges in non merged part 
      }

      ng.remove_isolated(v2);
      ng.remove_isolated(v1);

      char *cdstr = ng.col_graph_to_graph(size);
      // when problem is symmetric then red edges is not more than blue ones 
      // when equal then a word (canonic) is not greater (lexically)         
      if (symmetric && (edges[2] <= edges[3])) { 
        char *cdstr_2 = col_graph(&ng).col_graph_to_graph(size);
        if (edges[2] < edges[3] || strcmp(cdstr, cdstr_2) > 0) {
          FREES(cdstr);
          cdstr = cdstr_2;

          int i = edges[2];
          edges[2] = edges[3];
          edges[3] = i;
        }
        else FREES(cdstr_2);
      }
      graph_create++;
      i = mem_avail;
      inserttext((void**)graph_dict+(edges[2]*(max_edges+1)+edges[3]), cdstr, -1);
      FREES(cdstr);
      if (i == mem_avail) graph_stop++;
      if (mem_avail < 0) save_dictionary(); // dictionary is to big - dump it into disk       

      v--;
      if (v < 2) break; // otherwise an error when g1->==2
    }
    if (assign[v] < g2->n) {  // vertex v recently had lower number or was not assigned - go further 
      still_good = 0;
      if (assign[v] >= 0) {    // release vertex in g2
        i = assign[v];
        used[i] = 0; 
        //counting edges
        for (r = 0; r < nr_colours+2; r++) {
          g1->edges[r] += g1->degree[r][v];
          g2->edges[r] += g2->degree[r][i];
        }
        for (j = 0; j < v; j++) {
          if (assign[j] < g2->n) {
            g2->degree[g2->g[i][assign[j]]][i]++;
            g2->degree[g2->g[i][assign[j]]][assign[j]]++;
            if (g1->g[v][j] == 0) edges3[g2->g[i][assign[j]]]--;
          }
          g1->degree[g1->g[v][j]][v]++;
          g1->degree[g1->g[v][j]][j]++;
          edges3[g1->g[v][j]]--;
        }
      }
      for (i = assign[v]+1; i < g2->n; i++)
        if (!used[i]) {
          still_good = 1;
          for (j = 0; (j < v) && still_good; j++)
            if (assign[j] < g2->n)
              if ((g2->g[i][assign[j]] != 0) && (g1->g[v][j] != 0) && (g2->g[i][assign[j]] != g1->g[v][j]))
                still_good = 0;

          if (still_good) {
            used[i] = 1;
            assign[v] = i;
            //counting edges
            for (j = 0; j < v; j++) {
              if (assign[j] < g2->n) {
                g2->degree[g2->g[i][assign[j]]][i]--;
                g2->degree[g2->g[i][assign[j]]][assign[j]]--;
                if (g1->g[v][j] == 0) edges3[g2->g[i][assign[j]]]++;
              }
              g1->degree[g1->g[v][j]][v]--;
              g1->degree[g1->g[v][j]][j]--;
              edges3[g1->g[v][j]]++;
            }
            for (r = 0; r < nr_colours+2; r++) {
              g1->edges[r] -= g1->degree[r][v];
              g2->edges[r] -= g2->degree[r][i];
            }
            if (est_edges(g1, g2, v) <= max_edges)
              if (v >= limit || estimate_edges(g1, g2, v+1) <= max_edges) v++; 
              else est2_stop++;
            else est1_stop++;
            break;
          }
        }
      if (!still_good) { // vertex v wasn't assigned - assign as a new vertex
        assign[v] = size++;
        //counting edges
        for (j = 0; j < v; j++) {
          g1->degree[g1->g[v][j]][v]--;
          g1->degree[g1->g[v][j]][j]--;
          edges3[g1->g[v][j]]++;
        }
        for (r = 0; r < nr_colours+2; r++) 
          g1->edges[r] -= g1->degree[r][v];
        if (est_edges(g1, g2, v) <= max_edges)  
          if (v >= limit || estimate_edges(g1, g2, v+1) <= max_edges) v++; 
          else est2_stop++;
        else est1_stop++;
      }
    }
    else {  // vertex v had high number (outside g2) - release v
      //counting edges
      for (r = 0; r < nr_colours+2; r++) 
        g1->edges[r] += g1->degree[r][v];
      for (j = 0; j < v; j++) {
        g1->degree[g1->g[v][j]][v]++;
        g1->degree[g1->g[v][j]][j]++;
        edges3[g1->g[v][j]]--;
      }
      size--;
      assign[v] = -1;
      v--;
    }  
  } while (v > 1);
}

void combine(col_graph * g1, col_graph * g2) {
  // create new congigurations from g1 and g2
  int i, j, k, m, r, s, s00, s01, s10, s11;
  char *p1 = g1->descr, *p2;

  while (*p1) {
    i = *(p1++) - '0';
    j = *(p1++) - '0'; //  if (g1->g[i][j] >= 2) - it should be fulfilled 
    g1->degree[g1->g[i][j]][i]--;
    g1->degree[g1->g[i][j]][j]--;
    g1->edges[g1->g[i][j]]--;        
    for (r = 0; r < nr_colours+2; r++) 
      g1->edges[r] -= g1->degree[r][i] + g1->degree[r][j];
    if (i != 0)  g1->switch_vertices(0, i);    // permutation of vertices in which i,j are at the begining 
    if (j != 1)  g1->switch_vertices(1, j);    // done by two replacements (0<->i i 1<->j)

    p2 = g2->descr;
    while (*p2) {
      k = *(p2++) - '0';
      m = *(p2++) - '0'; // if (g2->g[k][m] >= 2) -  it should be fulfilled
      if (g1->g[0][1] != g2->g[k][m]) { // does not work for more than 2 colours
        g2->degree[g2->g[k][m]][k]--;
        g2->degree[g2->g[k][m]][m]--;
        g2->edges[g2->g[k][m]]--;
        for (r = 0; r < nr_colours+2; r++) 
          g2->edges[r] -= g2->degree[r][k] + g2->degree[r][m];

        s = s00 = s01 = s10 = s11 = 0;        // we count edges to check whether it is possible  
        for (r = 2; r < nr_colours+2; r++) {  // to merge small enough graph
          s += max(g1->edges[r], g2->edges[r]);
          s00 += max(g1->degree[r][0], g2->degree[r][k]);
          s11 += max(g1->degree[r][1], g2->degree[r][m]);
          s01 += max(g1->degree[r][0], g2->degree[r][m]);
          s10 += max(g1->degree[r][1], g2->degree[r][k]);
        }
        est1_count += 2;
        if (s+s00+s11 <= max_edges) combine(g1,g2,0,1,k,m); else est1_stop++;
        if (s+s01+s10 <= max_edges) combine(g1,g2,1,0,k,m); else est1_stop++;                              
        for (r = 0; r < nr_colours+2; r++) 
          g2->edges[r] += g2->degree[r][k] + g2->degree[r][m];
        g2->degree[g2->g[k][m]][k]++;
        g2->degree[g2->g[k][m]][m]++;
        g2->edges[g2->g[k][m]]++;
      }
    }
    if (j != 1) g1->switch_vertices(1, j);    // back to identity 
    if (i != 0)  g1->switch_vertices(0, i);    // replacements in the reverse order (1<->j i 0<->i) 

    for (r = 0; r < nr_colours+2; r++) 
      g1->edges[r] += g1->degree[r][i] + g1->degree[r][j];
    g1->degree[g1->g[i][j]][i]++;
    g1->degree[g1->g[i][j]][j]++;
    g1->edges[g1->g[i][j]]++;
  }
}
