///////////////////////////////////////////////////
// Ramsey On-line game
// Author: Przemek Gordinowicz 
//         Pawel Pralat 
///////////////////////////////////////////////////
//
//  Part of the main file (ramsey.cpp) including definition of a class of a graph with
//  coloured edges, to maintan ramsey configurations and converting it to the nauty format
//
///////////////////////////////////////////////////
class col_graph
{
public:
  int n;
  int **g;
  int edges[nr_colours+2];
  int *(degree[nr_colours+2]);
  char *descr;
// 0 - not determined
// 1 - non-edge
// 2 - first colour
// 3 - second colour
// ...
// nr_colours+1 - last colour
  static void free_mem();
  col_graph(int size);
  col_graph(graph *_g, int size);
  col_graph(col_graph *cg);
  ~col_graph();
  void print_graph();
  void remove_isolated(int v);
  void switch_vertices(int u, int v);
  char* col_graph_to_graph(int size);
  void calculate_degree();
  void sort_by_degree();
  void create_descr();
  void col_to_zero(int c);
};

DYNALLSTAT(graph,_g, _g_sz);    // global auxiliary variables to col_graph_to_graph
DYNALLSTAT(graph,_h,_h_sz);     // created to not allocate memory repeately
DYNALLSTAT(char,_fmt,_fmt_sz);

void col_graph::free_mem() {
  DYNFREE(_fmt, _fmt_sz);
  DYNFREE(_h, _h_sz);
  DYNFREE(_g, _g_sz);
}

// create empty graph on n vertices
col_graph::col_graph(int size) {
  int i;

  n = size;
  g = new int* [n];
  g[0] = new int [n*n];
  memset(g[0], 0, n*n * sizeof(int));
  for (i = 1 ; i < n ; i++)
    g[i] = g[i-1]+n;
  degree[0] = NULL;
  descr = NULL;
}

//creates graph with coloured edges (col_graph) from nauty graph (having "size/levels"
//coloured vertices in each layer 
col_graph::col_graph(graph *_g, int size) {
  n = size/levels;
  g = new int* [n];
  g[0] = new int [n*n];
  degree[0] = NULL;
  if (description) {
    descr = new char[strlen(description)+1];
    strcpy(descr, description);
  }
  else descr = NULL;

  int i, j, r, x;
  int _m;  
  int *perm = new int[n];
  set *_gv;
    
  memset(g[0], 0, n*n * sizeof(int));
  perm[0] = 0;
  for (i = 1 ; i < n ; i++) {
    g[i] = g[i-1]+n;
      perm[i] = i;
  }
  _m = (size + WORDSIZE - 1) / WORDSIZE;

  x = 1;
  for (r = 0; r < n*levels; r+=n) {
    for (i = 0; i < n; i++) {
      _gv = GRAPHROW(_g, perm[i], _m);
    for (j = i+1; j < n; j++)
      if (ISELEMENT(_gv, perm[j]))
        g[n-i-1][n-j-1] = (g[n-j-1][n-i-1] |= x);
    // position of the vertex corresponding to i-th in next layer 
    if (r < n*(levels-1))
    for (j = r + n; j < r + 2*n; j++)
      if (ISELEMENT(_gv, j)) {
        perm[i] = j;
        break;
      }
    }      
    x = x << 1;
  }
  delete[] perm;
}


// create symmetric copy of graph cg (exchange colours according to "perm") - works only for 2 colours
col_graph::col_graph(col_graph *cg) {
  int i,j;
  int perm[] = {0,1,3,2};

  n = cg->n;
  g = new int* [n];
  g[0] = new int [n*n];
  for (i = 1 ; i < n ; i++)
    g[i] = g[i-1]+n;

  for (i = 0 ; i < n ; i++) 
    for (j = 0 ; j < n ; j++)
      g[i][j] = perm[cg->g[i][j]];

  if (cg->degree[0]) {
    degree[0] = new int[100*(nr_colours+2)];
    for (j = 1; j < nr_colours+2; j++)
      degree[j] = degree[j-1]+100;
    for (j = 0; j < nr_colours+2; j++) {
      memcpy(degree[perm[j]], cg->degree[j], 100*sizeof(int));      
      edges[perm[j]] = cg->edges[j];
    }
  }
  else degree[0] = NULL;

  if (cg->descr) {
    descr = new char[strlen(cg->descr)+1];
    strcpy(descr, cg->descr);
  }
  else descr = NULL;
}

col_graph::~col_graph() {
  if (degree[0]) delete[] degree[0];
  if (descr) delete[] descr; 
  delete[] g[0];
  delete[] g;
}

void col_graph::print_graph() {
  int i,j;
  for (i = 0 ; i < n ; i++) {
    for (j = 0 ; j < i ; j++)
      printf("%d", g[i][j]);
    printf("\n");
  }
}

// if vertex v has no incident coloured edges it is removed 
// We know that there is some coloured edge
void col_graph::remove_isolated(int v) {
  int i, d;

  d = 0;
  for (i = 0; i < n; i++)
    if (g[i][v] > 1) d++;

  if (!d) { //remove
    n--;
    for (i = 0 ; i <= n; i++)
      memcpy(g[i]+v, g[i]+(v+1), (n-v)*sizeof(int));
    if (!v) memcpy(g[0], g[1], n*sizeof(int));
    else g[v] = g[v+1];
    for (i = v+1 ; i < n; i++)
      g[i] = g[i+1];
  }
}

// permutes vertices u and v - to release memory correctly it is important to rollback at final step 
// Remark: degree[0] != NULL
void col_graph::switch_vertices(int u, int v) {
  int i, x;
  int *row;

  // exchange rows 
  row = g[u]; g[u] = g[v]; g[v] = row;
  // exchange cols
  for (i = 0; i <n; i++) {
    x = g[i][u]; g[i][u] = g[i][v]; g[i][v] = x;
  }
  for (i = 0; i < nr_colours+2; i++) {
    x = degree[i][u]; degree[i][u] = degree[i][v]; degree[i][v] = x;
  }
}

//converts graph with coloured edges (col_graph) into nauty graph ("size" coloured vertices in 
//each layer), then converts it to isomorphic canonical form and return as text 
char* col_graph::col_graph_to_graph(int size) {
  int _m, _n;
  int i, j, k, r;
  char *_cdstr;
  set *_gv;

  _n = n * levels;
  _m = (_n + WORDSIZE - 1) / WORDSIZE;
  
  DYNALLOC2(graph,_g,_g_sz,_m,_n,"malloc");
  DYNALLOC2(graph,_h,_h_sz,_m,_n,"malloc");
  DYNALLOC1(char, _fmt, _fmt_sz, _n+1,"malloc"); 

  //conversion to _g, see Nauty User Guide p. 25;
  for (i = 0; i < _n; i++) {
    _gv = GRAPHROW(_g, i, _m);
    EMPTYSET(_gv, _m);
    // connect layers
    if (i < _n - n)  ADDELEMENT(_gv, i+n);
    if (i >= n) ADDELEMENT(_gv, i-n);
  }
  for (i = 0; i < n; i++)
    for (j = 0; j < i; j++) 
      if (k = g[i][j]) {
        r = 0;
        do {
          if (k % 2) {
            _gv = GRAPHROW(_g, i + r, _m);
            ADDELEMENT(_gv, j + r);
            _gv = GRAPHROW(_g, j + r, _m);
            ADDELEMENT(_gv, i + r);
          }
          r += n;
        } while (k = k >> 1);
      }
  // colours of layers
  if (!size) for (i = 0; i < levels; i++) 
    for (j = 0; j < n; j++)
      _fmt[i*n+j] = 2*(char)i + (used[j] ? '0' : '1');
  else for (i = 0; i < levels; i++) 
    for (j = 0; j < n; j++)
      _fmt[i*n+j] = (char)i + '0';
  
  _fmt[_n] = 0;
  // canonical form
  fcanonise(_g, _m, _n, _h, _fmt, false);
  _cdstr = pg_ntog6(_h, _m, _n);

  return _cdstr;
}

void col_graph::calculate_degree() { // computes (coloured) degrees and number of edges 
  int i, j;

  degree[0] = new int[100*(nr_colours+2)];
  memset(degree[0], 0, 100*(nr_colours+2)*sizeof(int));
  for (j = 1 ; j < nr_colours+2 ; j++)
    degree[j] = degree[j-1]+100;
  
  for (i = 0; i < n; i++)
    for (j = i+1; j < n; j++) {
      degree[g[i][j]][i]++;
      degree[g[i][j]][j]++;
    }
  memset(edges, 0, (nr_colours+2)*sizeof(int));
  for (j = 0; j < nr_colours+2; j++) {
    for (i = 0; i < n; i++)
      edges[j] += degree[j][i];
    edges[j] = edges[j] / 2;
  }
}

void col_graph::sort_by_degree() { // sorting vertices descending according to degree (2 colours) 
  int *score = new int[n];
  int *perm = new int[n];
  int i, j, k;
  
  for (i = 0; i < n; i++) {
    perm[i] = i;
    score[i] = ((degree[2][i] < degree[3][i]) ? 3*(int)degree[3][i]+2*degree[2][i] : 3*(int)degree[2][i]+2*degree[3][i]) + degree[1][i];
  }

  //insertion sort 
  for (i = 1; i < n; i++) {
    k = perm[i];
    for (j = i; j > 0 && score[k] > score[perm[j-1]]; j--)
      perm[j] = perm[j-1];
      perm[j] = k;
    }
  
  //application of sorting permutation 
  int **h = new int* [n];
    h[0] = new int [n*n];
  for (i = 1 ; i < n ; i++)
    h[i] = h[i-1]+n;
  for (i = 0 ; i < n ; i++) 
    for (j = 0 ; j < n ; j++)
      h[i][j] = g[perm[i]][perm[j]];
  delete[] g[0];
  delete[] g;
  g = h;
  
  for (j = 0; j < nr_colours+2; j++) {
    for (i = 0; i < n; i++)
      score[i] = degree[j][perm[i]];
    for (i = 0; i < n; i++)
      degree[j][i] = score[i];
  }

  if (descr) {
    char *p = descr;
    for (i = 0; i < n; i++)
      score[perm[i]] = i;
    while (*p) {
      i = (score[*p-'0']); 
      j = (score[*(p+1)-'0']); 
      if (i < j) {
        *(p++) = char(i) + '0';
        *(p++) = char(j) + '0';
      }
      else {
        *(p++) = char(j) + '0';
        *(p++) = char(i) + '0';
      }
    }
  }

  delete[] perm;
  delete[] score;
}

void col_graph::create_descr() { //creates description/list of coloured edges 
  if (!descr) {
    int i, j;
    char *p;
    i = edges[2] + edges[3];
    descr = new char[2*i+1];
    p = descr;

    for (i = 0; i < n; i++)
      for (j = i+1; j < n; j++)
        if (g[i][j] > NON_EDGE) {
          *(p++) = '0'+ (char)i;
          *(p++) = '0'+ (char)j;
        }
    *p = 0;
  }
}

void col_graph::col_to_zero(int c) { // exchange colours of edges c -> 0 (NON_EDGE)
  int i, j;
  for (i = 0; i < n; i++)
    for (j = i+1; j < n; j++)
      if (g[i][j] == c) 
        g[i][j] = g[j][i] = 0;
}
